/*! @file
  @brief Control module for Wi.232 module

  The Soma Project

  Copyright 2007 Rachel Pasetes, John Burr, Sean Eves Knudsen,
  Thomas Gerdes, Andrew Parra, Erik Pasternak

  This file is part of The Soma Project.

  The Soma Project is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  The Soma Project is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with The Soma Project; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA



  This file and its header provide various macros for bitmasking, pin
  definitions, and control of LEDs, motors, and inline functions for
  IO initilization */

#ifdef OBSERVE
#include <Windows.h>
#endif
#include "Packet.h"
#include "../share/util.h"
#include "../share/usart.h"

#ifdef DEBUG
#define dprintf(...) printf(__VA_ARGS__)
#else
#define dprintf(...) 
#endif

#define NET_SIZE 20           //max expected robots in swarm up to 256
#define MAXPACKETIDS 20

uint8_t PacketID;


uint8_t packetArray[1024];
uint16_t frontPacket;         //index of frontPacket
uint16_t backPacket;          //index of backPacket
uint8_t numPackets;
static uint8_t ListOfPacketIDS[MAXPACKETIDS][2];  //initialized to zero.  No packets sent may be packet 0
uint8_t packetIDCounter = 0;

int8_t RobotID = 0;
uint8_t machineState = 0;     //0: network init; 1:mapping
uint8_t isLeader = 0;
uint8_t leader;
uint8_t leaderChange = 0;
uint8_t justaddbool = 0;


uint8_t heartBeatBool = 0;

static uint8_t network[NET_SIZE];
uint8_t networkElements = 0;

#ifdef OBSERVE
DWORD WINAPI PacketThread(LPVOID lpParam) {
   while(1) {
      gatherPacket();
      Sleep(1);
   }
}
#endif

int Packet_Init( void ) {
#ifdef OBSERVE
   long idPacket;
#endif
   PacketID = 30;
   RobotID = getUserAddrSrc();
   if(RobotID < 0) return 1;
   backPacket  = 0;
   frontPacket = 0;
   numPackets  = 0;

#ifdef OBSERVE
   CreateThread(NULL, 0, PacketThread, NULL, 0, &idPacket);
#endif

   networkElements = 1;
   network[0] = RobotID;
   machineState = 0;
   return 0;
}

uint8_t bufferAvailable( uint8_t sizeOfPacket ) {
   uint8_t output = 0;
   int roomInBuffer = 0;

   if (frontPacket == backPacket) {
      if( numPackets == 0)
         roomInBuffer = 1023;
      else
         roomInBuffer = 0;
   }
   else if (frontPacket>backPacket) {
      roomInBuffer = 1023-(frontPacket - backPacket);
   }
   else {
      roomInBuffer = (1023- (1023-backPacket)) + frontPacket;
   }

   if (sizeOfPacket<=roomInBuffer)
      output = 1;
   else
      output = 0;    

   return output;
}

void enqueuePacket(packet tempPacket)
{
   int i = 0;

   if(bufferAvailable(tempPacket.length + 6))
   {
      packetArray[frontPacket] = tempPacket.length;
      frontPacket = (frontPacket+1)%1024;

      packetArray[frontPacket] = tempPacket.type;
      frontPacket = (frontPacket+1)%1024;

      packetArray[frontPacket] = tempPacket.RoboID;
      frontPacket = (frontPacket+1)%1024;

      packetArray[frontPacket] = tempPacket.PacketID;
      frontPacket = (frontPacket+1)%1024;

      for(i=0; i<tempPacket.length; i++) {
         packetArray[frontPacket] = tempPacket.Data[i];
         frontPacket = (frontPacket+1)%1024;
      }               

      packetArray[frontPacket] = (tempPacket.CheckSum>>8);
      frontPacket = (frontPacket+1)%1024;

      packetArray[frontPacket] = (tempPacket.CheckSum & 0xFF);
      frontPacket = (frontPacket+1)%1024;

      numPackets++;
   }

   return;
}

packet dequeuePacket( void )
{
   int i = 0;
   packet tempPacket;

   if(numPackets>0) {
      tempPacket.length = packetArray[backPacket];
      backPacket = (backPacket+1)%1024;

      tempPacket.type = packetArray[backPacket];
      backPacket = (backPacket+1)%1024;

      tempPacket.RoboID = packetArray[backPacket];
      backPacket = (backPacket+1)%1024;

      tempPacket.PacketID = packetArray[backPacket];
      backPacket = (backPacket+1)%1024;

      for(i=0; i<tempPacket.length; i++)
      {
         tempPacket.Data[i] = packetArray[backPacket];
         backPacket = (backPacket+1)%1024;
      }               

      tempPacket.CheckSum = (packetArray[backPacket]<<8);
      backPacket = (backPacket+1)%1024;
      tempPacket.CheckSum|= (packetArray[backPacket]);
      backPacket = (backPacket+1)%1024;

      numPackets--;
   }

   return tempPacket; 
}


uint16_t generateDataSum(uint8_t* data, uint8_t length)
{
   uint8_t i   = 0;
   int16_t dataSum = 0;

   for(i=0; i<length; i++) {
      dataSum+= data[i];
   }
   return dataSum;
}

uint16_t generateSum(packet receivedPacket)
{
   uint16_t sum = 0;

   sum += receivedPacket.length + receivedPacket.type + 
          receivedPacket.RoboID + receivedPacket.PacketID;
   sum += generateDataSum(receivedPacket.Data, receivedPacket.length);

   return sum;
}


uint8_t checkSum(packet pkt) {
   return (generateSum(pkt) == (pkt.CheckSum));
}




void sendPacket(uint8_t length, uint8_t type, uint8_t* data) {
   int i;
   packet transmissionPacket;
   transmissionPacket.length = length;
   transmissionPacket.type = type;
   transmissionPacket.RoboID = RobotID;   //actual code
   transmissionPacket.PacketID = PacketID;
   for(i=0; i<transmissionPacket.length; i++)
      transmissionPacket.Data[i] = data[i];

   transmissionPacket.CheckSum = generateSum(transmissionPacket);


   RF_Transmit(transmissionPacket.length);
   RF_Transmit(transmissionPacket.type);
   RF_Transmit(transmissionPacket.RoboID);
   RF_Transmit(transmissionPacket.PacketID);

   for(i=0; i<transmissionPacket.length; i++)
      RF_Transmit(transmissionPacket.Data[i]);  

   RF_Transmit(transmissionPacket.CheckSum>>8);
   RF_Transmit(transmissionPacket.CheckSum);

   return;
}


void incPacketID ( void )
{
   PacketID++;
   if(PacketID == 0)
      PacketID++;

   return;
}

uint8_t hasPackets( void )
{
   return (numPackets>0)? 1: 0;
}

void gatherPacket( void )
{
   uint8_t error = 0;
   uint8_t i; 
   uint8_t invalidPacket = 0;

   packet pkt;
   pkt.length   = RF_Receive(&error);
   if (error != 0) return; 
   pkt.type = RF_Receive(&error);
   if (error != 0) return;
   pkt.RoboID   = RF_Receive(&error);
   if (error != 0) return;
   pkt.PacketID = RF_Receive(&error);
   if (error != 0) return;      

   for(i=0; i<pkt.length; i++) {
      pkt.Data[i] = RF_Receive(&error);
      if (error != 0) return;  
   }

   //most significant byte of checkSum
   pkt.CheckSum = RF_Receive(&error)<<8; 
   if (error != 0) return;
   //least significant byte of checkSum
   pkt.CheckSum |= RF_Receive(&error);
   if (error != 0) return;

   //if you've made it this far, you've got all your data!
   if (checkSum(pkt)) {
      for(i=0; i<MAXPACKETIDS; i++) {
         if((ListOfPacketIDS[i][0] == pkt.PacketID) && (ListOfPacketIDS[i][1] == pkt.RoboID)) {
            invalidPacket = 1;
            break;
         }
      }

      if (!invalidPacket) {
         enqueuePacket(pkt);
         ListOfPacketIDS[packetIDCounter][0] = pkt.PacketID;
         ListOfPacketIDS[packetIDCounter][1] = pkt.RoboID;
         packetIDCounter = (packetIDCounter+1)%MAXPACKETIDS;
      }

//      dprintf("PacketIDCounter=%d  ;\n\r", packetIDCounter);

   }
   return;
}


/***************************************************************************/


void packetHandle (void) {

   packet unhandled;

   while (hasPackets()) {

      unhandled = dequeuePacket();
      //dprintf("Packet Received, of type: %d\n\r", unhandled.type);
      switch( unhandled.type )
      {
         //      case MAP:
         //         mapHandle(unhandled);
         //       break;
         //      case REQ_MAP:
         //         mapRequested(unhandled);
         //       break;
         case HBEAT:
            heartBeatHandle();
            break;
         //      case TOFINIT:
         //         timeOfFlightInit(unhandled);
         //       break;
         case BEGINIT:
            addToNet(unhandled);
            break;
         case PKTACK:
            if(isLeader)
               changeLeaderAck(unhandled);
            break;
         case TOKPASS:
            changeLeaderRequestHandle(unhandled);
            break;
         case START:
            startUp();
            break;
         case NET_IS:
   	        leadersNetwork(unhandled);   
            break;
         case DROPPED:
            handleDropped(unhandled);
            break;
         case REQNET:
            if(isLeader)
               addNewBotToNet(unhandled);
            break;
         case JUSTADDED:
            justaddbool = 1;
            leadersNetwork(unhandled);
            machineState = 1;
            dprintf("Machine State = 1\n\r");
            break;
//			case MAP:	
//					if(unhandled.Data[0] == 'w')
//						machineState = 1;
//					break;
         default :
            break;
      }
   }
   return;
}



/* populates network array with zeroes, establishes robot on network and tells other robots */
void initNetwork (void) {
   int i = 0;

   networkElements = 0;

   for(i=0; i<NET_SIZE; i++) {
      network[i] = 0;
   }
 
   network[0] = RobotID;
   networkElements++;

   return;
}

/* establishes a robot as the leader */
void initLeader (void) {
   int i;
   int lowest;

   lowest = network[0];
   for(i=0; i < networkElements; i++) {
      if( network[i]<lowest )
	     lowest = network[i];
   }

   if (RobotID == lowest)
      isLeader = 1;

   return;
}

/* leader sends his network to all robots */
void leaderConfirmNetwork (void) {

   if(isLeader) {
      sendPacket(networkElements, NET_IS, network);
      sendPacket(networkElements, NET_IS, network);
      sendPacket(networkElements, NET_IS, network);
      incPacketID();
      dprintf("Confirmed Net!\n\r");
   }

   return;
}


/* when a leaders confirmed network packet is received a robot replaced his network list */
void leadersNetwork(packet unhandled) {

   int i = 0;
   uint8_t inNet = 0;

   isLeader = 0;
   if(justaddbool){
      dprintf("---------------Not Leader-------------\n\r");
   }
   for(i=0; i<NET_SIZE; i++){
      network[i] = 0;
   }
      
   for(i=0; i<unhandled.length; i++) {
      network[i] = unhandled.Data[i];
   }
   networkElements = unhandled.length;   

   for(i=0; i<unhandled.length; i++) {
      if(RobotID==unhandled.Data[i])
         inNet = 1;
   }

   if(!inNet) {
      network[networkElements] = RobotID;
      networkElements++;
      requestNet();
   }

   return;
}


void requestNet( void ) {
   sendPacket(networkElements, DROPPED, network);
   sendPacket(networkElements, DROPPED, network);
   sendPacket(networkElements, DROPPED, network);
   incPacketID();
   return;
}

void handleDropped( packet unhandled ) {
   if(isLeader)
      addToNet( unhandled );
   return;
}

void addNewBotToNet(packet unhandled){
   network[networkElements] = unhandled.Data[0];
   networkElements++;
   sendPacket(networkElements, JUSTADDED, network); 
   sendPacket(networkElements, JUSTADDED, network);
   sendPacket(networkElements, JUSTADDED, network);
   return;
}



/*   Verifies and adds a robot to the network  */
void addToNet(packet unhandled) {
   int i;
   int j;
   int networkMember;
   int temp = networkElements;

   for(i=0; i<unhandled.length; i++) {
      networkMember = 0;
      for(j=0; j<networkElements ; j++) {
         if(network[j] == unhandled.Data[i])
            networkMember = 1;
      }

      if(networkMember==0) {
         network[networkElements] = unhandled.Data[i];
         networkElements++;
      }
   }

   if(temp<networkElements){

      if((machineState == 1)&&(isLeader)) {
         sendPacket(networkElements, NET_IS, network);
      }
      else{
         sendPacket(networkElements, BEGINIT, network); 
         sendPacket(networkElements, BEGINIT, network);
         sendPacket(networkElements, BEGINIT, network);
         incPacketID();
         dprintf("BEGININIT!\n\r");
      }
   }	   
                                                  
   return;
}

/*   Removes a robots from the network  */
void rmFromNet( uint8_t RoboID )
{
   uint8_t i = 0;
   int elementFound = 0;

   if (RoboID != RobotID)
   {

      for(i=0; i<networkElements; i++){

         if(RoboID == network[i])
            elementFound = 1;
     
	     if(elementFound == 1){
            network[i] = network[i+1];
		    network[i+1] = 0;
	     }
      }

      if(elementFound == 1)
         networkElements--;
   }

   return;
}


/*   If a robot is requested to be the next leader, it acknowledges it, and becomes leader  */
void changeLeaderRequestHandle(packet unhandled)
{
   if(unhandled.Data[0] == RobotID){
      incPacketID();
      sendPacket(1, PKTACK, unhandled.Data);
      sendPacket(1, PKTACK, unhandled.Data);
      sendPacket(1, PKTACK, unhandled.Data);
      sendPacket(1, PKTACK, unhandled.Data);
	   incPacketID();
      isLeader = 1;
      dprintf("***************Is Leader*************\n\r");
	   leader = RobotID;
      dprintf("ACK SENT!\n\r");
      machineState = 5;
   }
//   else{
//      isLeader = 0;
//   }

   return;
}



/*   If a robot is the leader, request a change in leadership  */
void changeLeader( void )
{
   int i = 0;
   int temp = 0;
   uint8_t nextLeader;

   if(isLeader)
   {
      dprintf("networkElements = %d\n\r", networkElements);
      for(i=0; i<networkElements; i++)
	  {
         if( RobotID == network[i]){
		      temp = i;
            dprintf("temp = %d, network[i] = %d\n\r", temp, network[i]);
            break; 
         }
	  }

      if(networkElements == 1)
         nextLeader = RobotID;
      else if((temp+1)<networkElements)
      {
         nextLeader = network[temp+1];
         dprintf("In else if: nextLeader = %d\n\r", nextLeader);
      }
      else{
	     nextLeader = network[0];
        dprintf("In else: nextLeader = %d\n\r", nextLeader);	   
      }
      sendPacket(1, TOKPASS, &nextLeader);
      sendPacket(1, TOKPASS, &nextLeader);
      sendPacket(1, TOKPASS, &nextLeader);
      dprintf("sent TOKPASS!\n\r");
      incPacketID();
   }

   return;
}

/* if a robot receives an ack, leadership will be passed*/
void changeLeaderAck(packet unhandled)
{
   leader = unhandled.RoboID;
   isLeader = 0;
   dprintf("---------------Not Leader-------------\n\r");
   leaderChange = 1;
   dprintf("Got ACK!\n\r");

   return;
}

/* if a robot had assumed leadership with an ack do nothing, else remove it from the network */
void checkChangeLeader( void )
{
   uint8_t i = 0;
   uint8_t temp = 0;
   uint8_t nextLeader;

   if(leaderChange == 1){
     leaderChange = 0;
     machineState = 1;
     dprintf("In State 1\n\r");
   }
   else{
      for(i=0; i<networkElements; i++)
	  {
         if( RobotID == network[i])
		    temp = i;
	  }
      if((temp+1)<=networkElements)
         nextLeader = network[temp+1];
      else
	     nextLeader = network[0];

      rmFromNet( nextLeader );
      leaderConfirmNetwork ();
      machineState = 5;

   }

   return;
}





void heartBeatHandle(void) {
   heartBeatBool = 1;
   return;
}

void heartBeatClear(void) {
   heartBeatBool = 0;
   return;
}


void sendHB(void) {
   sendPacket(1, HBEAT, &network[0]);  //data is not significant
   sendPacket(1, HBEAT, &network[0]);
   sendPacket(1, HBEAT, &network[0]);
   incPacketID();
   dprintf("Sent HB!\n\r");

   return;
}

uint8_t getHB(void) {
   return heartBeatBool;
}


void joinNW(void) {

   uint8_t initNetTimer = 0;

   initNetwork();
   
   for(initNetTimer=0; initNetTimer<10; initNetTimer++) {
#ifdef DEBUG_RF
      _delay_sec(1);
#endif
      sendPacket(networkElements, BEGINIT, network);
      dprintf("inForLoop\n\r");
      gatherPacket();
      packetHandle();
   }
   incPacketID(); 


   return;
}


void startUp( void ) {
   machineState = 1;
   return;
}

void checkNW(void){
   sendPacket(1, REQNET, (uint8_t*)&RobotID);  //data is not significant
   sendPacket(1, REQNET, (uint8_t*)&RobotID);
   sendPacket(1, REQNET, (uint8_t*)&RobotID);
}

void RunNWStateMachine(void) {
   static int32_t i = 0; 
   static uint32_t k = 0;
   static uint32_t j = 0;
   static uint8_t counter = 0;
      switch(machineState){
         case 0:        //Initialization, am I in the net or not?
            checkNW();  //Check to see if there is a network out there
            machineState = 4;   //Check a bunch of the packets to see if there was a heartbeat
            break;
         case 1:        //Hearbeat, wait for HB
            if(getHB()){
               machineState = 2;
            }
            break;
         case 2:        //Got Hearbeat, we're synched
            heartBeatClear();
            machineState = 1;
            break;
/*         case 3:
            sendHB();
            break;  */    
         case 4:
            if(i == 1000){
               dprintf("In State 4\n\r");
               isLeader = 1;  //There was no ack recieved from another network, i am the leader of my own network
               dprintf("***************Is Leader*************\n\r");
               i = 0;
               machineState = 5;
            }
            else{
            #ifndef OBSERVE 
               _delay_us(1);
            #endif
               i++;
            }
            break;      
         case 5:              //I am the leader, send an HB.
            if(j==10000){   //After 1 second, send a heartbeat, but make sure you are still sending and recieving packets
               dprintf("In State 5\n\r");
               sendHB();
               j=0;
               machineState = 6; //I've just sent a heartbeat, now lets send the network
            }
            #ifndef OBSERVE 
            _delay_us(1);   
            #endif
            j++;         
            break;
         case 6:              //Im the leader, im going to send my network
            leaderConfirmNetwork();
            machineState = 7; //done sending network, now lets change the leader
            dprintf("NET_IS: ");
            for( counter = 0; counter<networkElements; counter++)
               dprintf("%d, ", network[counter]);
            dprintf("\n\r");
            break;
         case 7:              //Request for a new leader
            changeLeader();   
            machineState = 8;
            break;
         case 8:
            if(k==100){    //Wait for a little while before you check to see if there was an acknowledge sent or not
               dprintf("In State 8\n\r");
               checkChangeLeader(); //Check to see if an acknowledge was recieved for not
               k = 0;
            }
            #ifndef OBSERVE 
            _delay_us(1);
            #endif
            k++;
            break;


      }
      if(machineState != 5 && machineState != 4 && machineState != 8 && machineState != 1){
         dprintf("Current state = %d\n\r", machineState);
      }
}

#ifdef DEBUG_RF_REV2
int  main( void )
{

   asm("cli");
   USART0_Init(BAUD57600);
   RF_Init();
   Packet_Init();
   asm("sei");
   dprintf("RobotID is %d\n\r", RobotID);
   for(;;){
      RunNWStateMachine();
      gatherPacket();
      packetHandle();
   }
}
#endif
