/*
   The Soma Project

   Copyright 2007 Rachel Pasetes, John Burr, Sean Eves Knudsen,
                  Thomas Gerdes, Andrew Parra, Erik Pasternak

  This file is part of The Soma Project.

  The Soma Project is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  The Soma Project is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with The Soma Project; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/

/*!@file
   This file and its header provide various macros for bitmasking, pin
   definitions, and control of LEDs, motors, and inline functions for
   IO initilization
*/
#ifndef OBSERVE
   #define CMD_ON()   CLR_BIT( RF_PORT, CMD_BIT );
   #define CMD_OFF()  SET_BIT( RF_PORT, CMD_BIT );
   #define SEI()  asm("sei")
   #define CLI()  asm("cli")
#else
#include <Windows.h>
   #define SEI()
   #define CLI()
   #define _delay_us
#endif
#include "RF.h"
#include "../share/util.h"

#ifdef MCU_atmega1281
#define DD_REG              DDRD
#define CMD_BIT             DDD4
#define EX_BIT              DDD3 
#define BE_BIT              DDD5
#define CRI_BIT             DDD6

#define DD_REG_IN           DDRC
#define CTS_BIT             DDC0

#define RF_PORT             PORTD
#define RF_IN_PORT          PINC

#define URF_SetBaud     USART1_SetBaud
#define URF_Flush       USART1_Flush
#define URF_HasData     USART1_HasData
#define URF_SendingData USART1_SendingData
#define URF_Transmit    USART1_Transmit
#define URF_Receive     USART1_Receive

#else  // ATmega88

#define DD_REG              DDRD
#define CMD_BIT             DDD3
#define EX_BIT              DDD4
#define BE_BIT              DDD5
#define CRI_BIT             DDD6

#define DD_REG_IN           DDRC
#define CTS_BIT             DDC0

#define RF_PORT             PORTD
#define RF_IN_PORT          PINC

#define URF_SetBaud     USART0_SetBaud
#define URF_Flush       USART0_Flush
#define URF_HasData     USART0_HasData
#define URF_SendingData USART0_SendingData
#define URF_Transmit    USART0_Transmit
#define URF_Receive     USART0_Receive

#endif



/**** local definitions                           ****/ 
uint8_t currentBaud;

FILE *fRF;

const enum BaudRate RFBaudTouCBaud[] = {
   BAUD2400,
   BAUD9600,
   BAUD19200,
   BAUD38400,
   BAUD57600,
   BAUD115200 //,
   //BAUD10400,
   //BAUD31250
};

/**** RF_Init initializes microcontroller UART      ****/
/**** and I/O pins for RF transceiver communication ****/

inline void RF_Init ()
{
#ifndef OBSERVE
   SET_BIT( DD_REG, CMD_BIT );    //CMD is an output
   SET_BIT( RF_PORT, CMD_BIT );   //it's pulled high

   CLR_BIT( DD_REG, CTS_BIT );    //CTS is input
   CLR_BIT( DD_REG, EX_BIT );     //EX is input
   CLR_BIT( DD_REG, BE_BIT );     //BE is input

   //CLR_BIT( DD_REG, CRI_BIT );   //CRI is input
   //CLR_BIT( PINB, PF2 );   //EX's pullup is off
   //CLR_BIT( PINB, PF3 );   //BE's pullup is off
   //CLR_BIT( PINB, PF4 );   //CRI's pullup is off
   //SET_BIT( PORTF, PF0 );   //CTS's pullup is on

#ifdef MCU_atmega1281
   USART1_Init( BAUD57600);          //initializes RF UART
   fRF = fdevopen((void*)USART1_Transmit, (void*)USART1_Receive);
#else
   USART0_Init( BAUD57600);          //initializes RF UART
   fRF = fdevopen((void*)USART0_Transmit, (void*)USART0_Receive);
#endif

#endif
   return;
}

inline void    RF_Transmit( uint8_t data) {
#ifndef OBSERVE
   fputc(data, fRF);
   #define Sleep(x)
#else
   USART0_Transmit(data);
#endif
}
#define RETRIES   200
uint8_t RF_Receive ( uint8_t *err ) {
   int i = 0;
   *err = 0;
   while(i++ < RETRIES && !URF_HasData()) {
      Sleep(1);
   }
   if(!URF_HasData()) {
      return (*err = 1);
   } else {
   return URF_Receive();
   }
}

/**** returns the baud rate currently used        ****/
uint8_t getBaud() {
   return currentBaud;
}


/**** writes to the regNVUARTDATARATE (0x03)      ****/
/**** baud rate register of the Wi.232            ****/
/**** returns: 0x6 ACK, 0x15 NACK, 0x0 no command ****/
uint8_t setBaud( uint8_t baudRate )
{
   uint8_t ackBaudChange = 0;
   uint8_t counter = 0;

   if ( (baudRate <= 0x7) )
   {
      CLI();
      URF_Flush();
      CMD_ON();
      SEI();

      fputc(0xFF, fRF);
      fputc(0x02, fRF);
      fputc(UARTDataRate, fRF);
      fputc(baudRate, fRF);

      while(URF_SendingData()) ;
      delayLoop();

      do
      {
         if(URF_HasData())
            ackBaudChange = fgetc(fRF);  
      }while((++counter<10)&&(ackBaudChange==0));

      if ((ackBaudChange!=0))
      {
         currentBaud = baudRate;
         URF_SetBaud(RFBaudTouCBaud[baudRate]);     
      }

      _delay_us(40);
      CMD_OFF();
   }
   return ackBaudChange;
}


/**** writes to the reg (0x52)                    ****/
/**** maxRetries register of the Wi.232            ****/

uint8_t getRetries( void )
{
   uint8_t ack = 0;
   uint8_t reg = 0;
   uint8_t value = 0;

   CLI();
   URF_Flush();
   CMD_ON();
   SEI();

   fputc(0xFF, fRF);
   fputc(0x02, fRF);
   fputc(0xFE, fRF);
   fputc(MaxRetries, fRF);

   while(URF_SendingData());
   delayLoop();
   ack = fgetc(fRF);
   reg = fgetc(fRF);
   value = fgetc(fRF);
   _delay_us(40);

   CMD_OFF();

   return value;
}



/**** writes to the reg (0x52)                    ****/
/**** maxRetries register of the Wi.232           ****/
/**** returns: 0x6 ACK, 0x15 NACK, 0x0 no command****/
uint8_t setRetries( uint8_t retries )
{
   uint8_t ackRetriesChange = 0;
   uint8_t counter = 0;

   if ( (retries<=26) )
   {
      CLI();
      URF_Flush();
      CMD_ON();
      SEI();

      fputc(0xFF, fRF);
      fputc(0x02, fRF);
      fputc(MaxRetries, fRF);
      fputc(retries, fRF);

      while(URF_SendingData()) ;
      delayLoop();

      do
      {
         if(URF_HasData())
            ackRetriesChange = fgetc(fRF);  
         counter++;
      }while((counter<10)&&(ackRetriesChange==0x0));

      _delay_us(40);
      CMD_OFF();
   }

   return ackRetriesChange;
}


/**** reads from the reg (0x4F)                   ****/
/**** Network Mode register of the Wi.232         ****/

uint8_t getMode( void )
{
   uint8_t ack = 0;
   uint8_t reg = 0;
   uint8_t value = 0;

   CLI();
   URF_Flush();
   CMD_ON();
   SEI();

   fputc(0xFF, fRF);
   fputc(0x02, fRF);
   fputc(0xFE, fRF);
   fputc(NetworkMode, fRF);

   while(URF_SendingData());
   delayLoop();
   ack = fgetc(fRF);
   reg = fgetc(fRF);
   value = fgetc(fRF);
   _delay_us(40);

   CMD_OFF();
   return value;
}



/**** writes to the reg (0x4F)                    ****/
/**** Network Mode register of the Wi.232         ****/
/**** returns: 0x6 ACK, 0x15 NACK, 0x0 no command ****/
uint8_t setMode( uint8_t mode )
{
   uint8_t ackModeChange = 0;   //selected because 5 will never be a value
   uint8_t counter = 0;

   if ( (mode == MAC)||(mode == User)||(mode == UserExt)||(mode == MACACK)||(mode == UserACK)||(mode == UserExtACK) )
   {
      CLI();
      URF_Flush();
      CMD_ON();
      SEI();

      fputc(0xFF, fRF);
      fputc(0x02, fRF);
      fputc(NetworkMode, fRF);
      fputc(mode, fRF);

      while(URF_SendingData()) ;
      delayLoop();

      do
      {
         if(URF_HasData())
            ackModeChange = fgetc(fRF);  
         counter++;
      }while((counter<10)&&(ackModeChange==0x0));

      _delay_us(40);
      CMD_OFF();
   }

   return ackModeChange;
}

/**** writes to the reg (0x4D)                    ****/
/**** power mode register of the Wi.232           ****/

uint8_t getPower( void )
{
   uint8_t ack = 0;
   uint8_t reg = 0;
   uint8_t value = 0;

   CLI();
   URF_Flush();
   CMD_ON();
   SEI();

   fputc(0xFF, fRF);
   fputc(0x02, fRF);
   fputc(0xFE, fRF);
   fputc(PowerMode, fRF);

   while(URF_SendingData());
   delayLoop();
   ack = fgetc(fRF);
   reg = fgetc(fRF);
   value = fgetc(fRF);
   _delay_us(40);

   CMD_OFF();
   return value;
}

/**** writes to the reg(0x4D)                     ****/
/**** power mode register of the Wi.232           ****/
/**** returns: 0x6 ACK, 0x15 NACK, 0x0 no command****/
uint8_t setPower( uint8_t power )
{
   uint8_t ackPowerChange = 0x0;   //selected because 5 will never be a value
   uint8_t counter = 0;

   if ( (power <= 0x3) )
   {
      CLI();
      URF_Flush();
      CMD_ON();
      SEI();

      fputc(0xFF, fRF);
      fputc(0x02, fRF);
      fputc(PowerMode, fRF);
      fputc(power, fRF);

      while(URF_SendingData()) ;
      delayLoop();

      do
      {
         if(URF_HasData())
            ackPowerChange = fgetc(fRF);  
         counter++;
      }while((counter<10)&&(ackPowerChange==0x0));

      _delay_us(20);
      CMD_OFF();
   }
   return ackPowerChange;
}



/**** writes to the reg (0x79)                    ****/
/**** exception register of the Wi.232            ****/

uint8_t getException( void )
{
   uint8_t ack = 0;
   uint8_t reg = 0;
   uint8_t value = 0;

   CLI();
   URF_Flush();
   CMD_ON();
   SEI();

   fputc(0xFF, fRF);
   fputc(0x02, fRF);
   fputc(0xFE, fRF);
   fputc(Exception, fRF);

   while(URF_SendingData());
   delayLoop();
   ack = fgetc(fRF);
   reg = fgetc(fRF);
   value = fgetc(fRF);
   _delay_us(20);

   CMD_OFF();
   return value;
}



/**** reads from the the regs (0x68-0x6B)         ****/
/**** exception register of the Wi.232            ****/

void getMACDest( uint8_t *value )
{
   uint8_t ack[4]   = {0,0,0,0};
   uint8_t reg[4]   = {DestGUID3, DestGUID2, DestGUID1, DestGUID0};
   uint8_t counter  = 0;

   for(counter = 0; counter<4; counter++)
   {

      CLI();
      URF_Flush();
      CMD_ON();
      SEI();
      fputc(0xFF, fRF);
      fputc(0x02, fRF);
      fputc(0xFE, fRF);
      fputc(reg[counter], fRF);

      while(URF_SendingData());
      delayLoop();
      ack[counter]   = fgetc(fRF);
      reg[counter]   = fgetc(fRF);
      value[counter] = fgetc(fRF);

      _delay_us(40);
      CMD_OFF();
   }
   return;
}

/**** writes to the reg(0x68-0x6B)                ****/
/**** exception register of the Wi.232            ****/
/**** returns: 0x6 ACK, 0x15 NACK, 0x0 no command****/
void setMACDest( uint8_t *value )
{
   uint8_t reg[4]   = {DestGUID3, DestGUID2, DestGUID1, DestGUID0};
   uint8_t ack[4]   = {0,0,0,0};
   uint8_t counter = 0;
   uint16_t i = 0;


   for(counter=0; counter<4; counter++)
   {

      CLI();
      URF_Flush();
      CMD_ON();
      SEI();       

      fputc(0xFF, fRF);
      fputc(0x02, fRF);
      fputc(reg[counter], fRF);
      fputc(value[counter], fRF);

      while(URF_SendingData()) ;
      delayLoop();

      i=0;
      do
      {
         if(URF_HasData())
            ack[counter] = fgetc(fRF);  
         i++;
      }while((i<10)&&(ack[counter]==0x0));

      _delay_us(20);
      CMD_OFF();
   }


   for(counter=0; counter<4; counter++)
   {
      value[counter] = ack[counter];
   }
   return;
}

/**** reads from reg(0x5D)                        ****/
/**** User-mode destination register of the Wi.232****/
/**** returns: 0x6 ACK, 0x15 NACK, 0x0 no command ****/
uint8_t getUserAddrDest( void )
{
   uint8_t ack = 0;
   uint8_t reg = 0;
   uint8_t value = 0;

   CLI();
   URF_Flush();
   while(URF_SendingData()) ;
   CMD_ON();
   SEI();

   fputc(0xFF, fRF);
   fputc(0x02, fRF);
   fputc(0xFE, fRF);
   fputc(UserDestID0, fRF);

   while(URF_SendingData());
   delayLoop();
   ack = fgetc(fRF);
   reg = fgetc(fRF);
   value = fgetc(fRF);
   _delay_us(20);

   CMD_OFF();

   return value;
}

/**** writes to the reg(0x5D)                     ****/
/**** User-mode destination register of the Wi.232****/
/**** returns: 0x6 ACK, 0x15 NACK, 0x0 no command ****/
uint8_t setUserAddrDest( uint8_t dest )
{
   uint8_t ackDestChange = 0x0;  
   uint8_t counter = 0;


   CLI();
   URF_Flush();
   CMD_ON();
   SEI();

   fputc(0xFF, fRF);
   fputc(0x02, fRF);
   fputc(UserDestID0, fRF);
   fputc(dest, fRF);

   while(URF_SendingData()) ;
   delayLoop();

   do
   {
      if(URF_HasData())
         ackDestChange = fgetc(fRF);  
      counter++;
   }while((counter<10)&&(ackDestChange==0x0));

   _delay_us(20);
   CMD_OFF();

   return ackDestChange;
}

/**** reads from reg(0x61)                        ****/
/**** User-mode source register of the Wi.232     ****/
/**** returns: 0x6 ACK, 0x15 NACK, 0x0 no command ****/
int8_t getUserAddrSrc( void )
{
   uint8_t ack = 0;
   uint8_t reg = 0;
   uint8_t value = 0;
   uint8_t err;

   CLI();
   URF_Flush();
   while(URF_SendingData()) ;
   CMD_ON();
   SEI();

   RF_Transmit(0xFF);
   RF_Transmit(0x02);
   RF_Transmit(0xFE);
   RF_Transmit(UserSourceID0);

   while(URF_SendingData());
   delayLoop();
   ack =   RF_Receive(&err); //if(err) return -1;
   reg =   RF_Receive(&err); //if(err) return -1;
   value = RF_Receive(&err); //if(err) return -1;

   _delay_us(20);
   CMD_OFF();

   return value;
}

/**** writes to the reg(0x61)                     ****/
/**** User-mode destination register of the Wi.232****/
/**** returns: 0x6 ACK, 0x15 NACK, 0x0 no command ****/
uint8_t setUserAddrSrc( uint8_t src )
{
   uint8_t ackSrcChange = 0x0;  
   uint8_t counter = 0;


   CLI();
   URF_Flush();
   CMD_ON();
   SEI();

   fputc(0xFF, fRF);
   fputc(0x02, fRF);
   fputc(UserSourceID0, fRF);
   fputc(src, fRF);

   while(URF_SendingData()) ;
   delayLoop();

   do
   {
      if(URF_HasData())
 	     ackSrcChange = fgetc(fRF);  
      counter++;
   }while((counter<10)&&(ackSrcChange==0x0));

   _delay_us(20);
   CMD_OFF();

   return ackSrcChange;
}


void delayLoop( void )
{
   int x;
   for(x=0; x<10; x++)    //delay
      _delay_us(100);
}

