/*! @file  RF.h nanomouse UART driver

   The Soma Project

   Copyright 2007 Rachel Pasetes, John Burr, Sean Eves Knudsen,
                  Thomas Gerdes, Andrew Parra, Erik Pasternak

   Code adapted from Nanomouse

   (c) 2006 Andrey Yurovsky
            andrey@thedotcommune.com

  This file is part of The Soma Project.

  The Soma Project is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  The Soma Project is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with The Soma Project; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/

#if !defined( RF_Driver_H )
#define RF_Driver_H

#ifndef OBSERVE //If compiling for the microprocessor
#include <avr/io.h>
#include <util/delay.h>
#endif

#include <stdio.h>
#include <string.h>
#include "../share/usart.h"

#include <stdint.h>



/* Acknowledge confirmation characters */
#define ACK                    0x06	   ///<Acknowledge.     Sent by module.
#define NACK                   0x15	   ///<No Acknowledge.  Sent by module.

/* Register */
#define HopTable               0x4B    ///<Valid values are 0-5
#define PowerMode              0x4D    ///<0,1,2,3 => Low,Mid-low, Mid-high, High
#define UARTDataRate           0x4E    ///<Baud Rate Reg.  See Baud rate values below
#define NetworkMode            0x4F    ///<See Network Mode values below.
#define XmitTimeout            0x50    ///<default x10; do not set below 0x02
#define MaxRetries             0x52    ///<Max retries: default 26
#define CRCCheck               0x53    ///<Error Checking.  0x01 to enable, 0x00 to disable.  
#define UARTMinXmit            0x54    ///<Default:64, Min:1, Max: 192.
#define Verbose                0x55    ///<Default:0x1. Responsible for start-up message.
#define CSMA                   0x56    ///<Default:0x1. Responsible for handling collisions.
#define OperatingMode          0x58    ///<Default: Awake(0); Sleep(1); Standby(2); Deep Sleep(3)
#define ACKOnWake              0x59    ///<Default:0x1. Acknowledge sent upon waking from sleep.
#define UserDestID3            0x5A    ///<Most significant byte of the user destination address.
#define UserDestID2            0x5B    ///<Byte 2 of the user destination address.
#define UserDestID1            0x5C    ///<Byte 1 of the user destination address.
#define UserDestID0            0x5D    ///<Least significant byte of the user destination address.
#define UserSourceID3          0x5E    ///<Most significant byte of the user source address.
#define UserSourceID2          0x5F    ///<Byte 2 of the user source address.
#define UserSourceID1          0x60    ///<Byte 1 of the user source address.
#define UserSourceID0          0x16    ///<Least significant byte of the user source address. 0x61V
#define UserIDMask3            0x62    ///<Most significant byte of the User ID Mask
#define UserIDMask2            0x63    ///<Byte 2 of the User ID Mask
#define UserIDMask1            0x64    ///<Byte 1 of the User ID Mask
#define UserIDMask0            0x65    ///<Least significant byte of the User ID Mask
#define DestGUID3              0x68    ///<Most siginicicant byte of MAC Destination GUID
#define DestGUID2              0x69    ///<Most siginicicant byte of MAC Destination GUID
#define DestGUID1              0x6A    ///<Most siginicicant byte of MAC Destination GUID
#define DestGUID0              0x6B    ///<Most siginicicant byte of MAC Destination GUID
#define ExceptionMask          0x6C    ///<Default:0xFF
#define CMDHaltsTraffic        0x6E    ///<Default:Disabled(0x00).  0x01 will buffer incoming data when CMD line is low.
#define ReceiveLNAMode         0x6F    ///<Default:AGC enabled(0x00)
#define Compatibility Mode     0x70    ///<Default:Disabled(0x00).
#define ReleaseNumber          0x78    ///<Hardcoded data:firmware and hardware platform
#define Exception              0x79    ///<Exception Register

/* Baud rates */                               
#define RFBAUD2400             0x0 		///<2400 
#define RFBAUD9600             0x1      ///<9600
#define RFBAUD10400            0x6      ///<10400
#define RFBAUD19200            0x2      ///<19200
#define RFBAUD38400            0x3      ///<38400
#define RFBAUD57600            0x4      ///<57600
#define RFBAUD115200           0x5      ///<115200
#define RFBAUD31250            0x7      ///<31250

/* Network Modes */
#define MAC                    0x04
#define User                   0x06
#define UserExt                0x07
#define MACACK                 0x14
#define UserACK                0x16
#define UserExtACK             0x17


/**** Function declarations ****/
inline void RF_Init ( void );

void    RF_Transmit( uint8_t data);
uint8_t RF_Receive ( uint8_t *err) ;

uint8_t getBaud( void );
uint8_t setBaud( uint8_t baudRate );
uint8_t getRetries( void );
uint8_t setRetries( uint8_t retries );
uint8_t getMode( void );
uint8_t setMode( uint8_t mode );
uint8_t getPower( void );
uint8_t setPower( uint8_t power );
uint8_t getException( void );

void getMACDest( uint8_t *value );
void setMACDest( uint8_t *value );

uint8_t getUserAddrDest( void );
uint8_t setUserAddrDest( uint8_t dest );
int8_t  getUserAddrSrc( void );
uint8_t setUserAddrDest( uint8_t dest );


void delayLoop( void );
#endif
