/*
   The Soma Project

   Copyright 2007 Rachel Pasetes, John Burr, Sean Eves Knudsen,
                  Thomas Gerdes, Andrew Parra, Erik Pasternak

  This file is part of The Soma Project.

  The Soma Project is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  The Soma Project is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with The Soma Project; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/

/*!@file
   @ingroup command
   Main program for Soma robots - function prototypes
*/


typedef struct pos {
   int x,    ///< horizontal position
   int y,    ///< vertical position
   int ang,  ///< angle, in degrees XXX?
   int cert  ///< how accurate we are sure of the position
} pos_t;

typedef struct robot {
   int id,
   pos_t pos,

} robot_t;
/*!@brief Maximum number of robots in memory is declared at compile time
   Adjust this value if there are ever more than 5 Max Robots.
 */
#define MAX_ROBOTS 5

/*!@brief Array holding info about all robots*/
robot_t Robots[MAX_ROBOTS];

/*!@brief Use the position of other robots to localize where this robot
   is located.
   
   This function signals to the localize board over SPI that it should 
   send Ultrasonic and Infrared.  It then should release control back 
   to the main program loop.  Data is then gathered from the SendLocalize
   function. 
 */
void SendLocalize( void );

/*!@brief Read position of the last robot who broadcasted from the 
 * localize board. 
 *
 * This function modifies the position of the previous broadcasting
 * robot in the Robot array.  This function should be called some
 * reasonable time after the Heartbeat should have occured.  The
 * function should poll SPI for a reasonable ammount of time before the
 * it times out and determines the other robot is not in line of sight.
 * If this is the case, the position is estimated based on known data,
 * and the certianty of the position is reduced.*/
void ReadLocalize( void );

/*!@brief */
void GetPosition( pos_t *pos);

/*!@brief Read distance reading from the distance ping sensor.
 *
 * This function gets called by UpdateMap.  Some range of results is
 * likely to be more accurate than others: we should probably determine
 * when a value is too far away to mean anything useful.
 * @returns the distance to the nearest obstcle in inches.*/
int ReadPingSensor(void);

/*!@brief Update map data based on sensor readings.
 *
 * This function should take data from the ping sensor and knowledge
 * about where all the robots are to update the map.  The uncertainty of
 * the positions scale the probability of accuracy of readings, as does
 * the value read (further distances are more likely to be unaccurate
 * than close measurements).*/
void UpdateMap( void );

/*!@brief Choose a new location to move to.
 *
 * Not only is the destination important, but the robot should actually
 * keep track of a path of waypoints to get to that location.  The
 * localize board and other robots need to only know about the immediate
 * (and maybe second) waypoints in order to opperate. This function
 * should also modify the path if there is an immediate obstacle in the
 * way of the current path.
 */
void PickDestination(void);


