/*
   The Soma Project

   Copyright 2007 Rachel Pasetes, John Burr, Sean Eves Knudsen,
                  Thomas Gerdes, Andrew Parra, Erik Pasternak

  This file is part of The Soma Project.

  The Soma Project is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  The Soma Project is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with The Soma Project; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/

/*!@file
   @ingroup command
   @brief Input Capture related functions
*/

#include "util.h"
#include <stdio.h>
#include <avr/io.h>
#include <avr/interrupt.h>
#include "usart.h"
#include "ic.h"
#include "oc.h"
#include "triglookup.h"
#include <math.h>

volatile unsigned int ICTimerValNew = 0;
volatile unsigned int ICTimerValOld;
volatile unsigned int ICTimerVal;


/*!@brief Initialize Input Capture 1

   This function initializes the Input Capture 1 to allow for
   interrupt driven input.  It is used for the ping sensors.
   The trigger is set to go off on a toggle high.
*/
inline void ic_Init( void )
{
   SET_BIT( TCCR3B, CS32 ); /* Clock select, no prescaling */
   CLR_BIT( TCCR3B, WGM32 ); /* Set the Waveform Generator to Normal */
   CLR_BIT( TCCR3B, WGM33 );
   CLR_BIT( TCCR3A, WGM31 );
   CLR_BIT( TCCR3A, WGM30 );
   CLR_BIT( DDRE, DDE7 ); /* Set Pin D4 as an input */
   SET_BIT( TIFR3, ICF3 ); /* Clear input capture flag */
   SET_BIT( TIMSK3, ICIE3 ); /* Enable Interrupt Input Capture Select */
   SET_BIT( TCCR3B, ICES3 ); /* Set to toggle high */
}

/*!@brief Interrupt service routine for TIMER1

   This function occurs whenever the TIMER1 input capture triggers.
   Sets the global variable of ICTimerVal to the time difference
   between rising and falling edge.
*/
ISR(TIMER3_CAPT_vect){

   ICTimerValOld = ICTimerValNew;  /* Put the old time in here */
   ICTimerValNew = ICR3;  /* Get the new time */

   if( BIT_IS_CLR( TCCR3B, ICES3 ) ) {  /* If we are on a falling edge */
      if( ICTimerValNew < ICTimerValOld )  /* If we have an overflow situation */
	     ICTimerVal = 0xFFFF - ICTimerValOld + ICTimerValNew; /* Account for the overflow */
      else
         ICTimerVal = ICTimerValNew - ICTimerValOld;
   }

   TOGGLE_BIT( TCCR3B, ICES3 );  /* Change the edge interrupt type */
}

/*!@brief Input Capture Return Distance in Inches
   @returns the distance from the sonar ping sensor to an object

   This function occurs whenever the TIMER1 input capture triggers.
   Sets the global variable of ICTimerVal to the time difference
   between rising and falling edge.
*/
int16_t ic_GetDistance( void ){
   return( ICTimerVal * INCHES_PER_CYCLE / 100 ); 
}

/*!@brief Input Capture Ping Sensor Data Return
   @returns A data structure with distance and angle of an object

   This function converts the data recieved from the ping sensor
   to data relevent towards the center of the robot.  This data
   includes the distance from the center of the robot to an
   onject, and the center of the robot and its angle relative to
   an object.
*/
ping_vector ic_PingSensorData( void ){
   int16_t distance_ping = ic_GetDistance();
   ping_vector data;
   data.distance = sqrt(distance_ping*distance_ping + 49 - 
          ((2 * distance_ping * 7 * cosd(270-oc_GetServoAngle())) >>10) );
   data.angle = sinInversed(((int32_t)distance_ping*sind(270-oc_GetServoAngle()))/data.distance);
   data.angle += 90;
   return data;
}

#ifdef DEBUG_IC

int main(void) {
   USART0_Init(BAUD57600);
   asm("sei");
   printf("Hello World\n\r");
   ic_Init();
   for(;;) { 
      getchar();
	  printf("%d   %d\n\r", ic_GetDistance(), INCHES_PER_CYCLE); 		
   }
}

#endif

#ifdef PINGSENSETEST
int main(void){
   ping_vector ping_data;
   int i,j;
   asm("sei");
   USART0_Init(BAUD57600);
   ic_Init();
   oc_Init();
   
   printf("Hello Andrew\n\r");
   for(;;){
      i = getchar();
      if(i == '1')
         oc_TurnServo(0);
      if(i == '2')
         oc_TurnServo(90);
      if(i == '3')
         oc_TurnServo(180);
      printf("%d, %d, ", oc_GetServoAngle(), ic_GetDistance());
      ping_data = ic_PingSensorData();
      printf("distance: %d angle: %d\n\r", ping_data.distance, ping_data.angle);
   }
}
#endif
