/*
   The Soma Project

   Copyright 2007 Rachel Pasetes, John Burr, Sean Eves Knudsen,
                  Thomas Gerdes, Andrew Parra, Erik Pasternak

  This file is part of The Soma Project.

  The Soma Project is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  The Soma Project is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with The Soma Project; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/

/*!@defgroup command Command module

   @file
   @ingroup command
   Main program for Soma robots
*/


/* Includes */

#include "RF.h"
#include "util.h"
#include "usart.h"
#include "spi.h"
#include "sram.h"
#include "oc.h"
#include "ic.h"
#include <stdio.h>
#include "localcom.h"
#include "triglookup.h"
#include "Packet.h"

void initAll(void) {{{
   asm("cli");
   SPI_Init(MASTER);
   RF_Init();
   ic_Init();
   oc_Init();
   USART0_Init( BAUD57600 );
   init_Chip_Select_Localize();
   Packet_Init();
   asm("sei");
   oc_TurnServo(90);
   _delay_ms(10);
}}}

void HandlePackets(void) {
   static int servo_angle=90;
   unsigned char command;
   packet pkt;
   gatherPacket();
   if(hasPackets()) {

      pkt = dequeuePacket();
      if(pkt.type == 1) { //TODO: change type to #define
         command = pkt.Data[0];
         if(command == '1'){
            servo_angle = servo_angle+10;
            if(servo_angle>180)
               servo_angle = 0;
            oc_TurnServo(servo_angle);

         }
         else{
            while(SPI_Transfer(SEND_COMMAND) != SPI_ACK) ;
            SPI_Transfer(command);
            _delay_ms(10);
         }
      }
   }
}

void doGetXY(uint8_t *data) {
   int i;
   while(SPI_Transfer(GETXY) != SPI_ACK) ; //keep sending till ack
   SPI_Transfer(0);
   SPI_Transfer(0);
   for(i = 0; i<6; i++){
      data[i] = SPI_Transfer(0);
   }
   if((data[0] & 0x70) != 0) 
      data[0] |= (0x80);
   else
      data[0] &= (0x7F);
}

#ifdef BLAH
int main(void){
   uint8_t data[9];
   ping_vector ping_data;

   initAll();
   for(;;){
      //HandlePackets();

      ping_data = ic_PingSensorData();
      if(ping_data.distance < 12 && ping_data.distance > 0){   //Obstacle detection
         printf("In if\n\r");
         SPI_Transfer(SEND_COMMAND);
         while(SPI_Transfer(SEND_COMMAND) != SPI_ACK);
         SPI_Transfer('a');
      } else {
         SPI_Transfer(SEND_COMMAND);
         while(SPI_Transfer(SEND_COMMAND) != SPI_ACK);
         SPI_Transfer('w');
      }
      _delay_ms(10);

      doGetXY(data);

      data[6] = ping_data.distance & 0xFF;
      data[7] = ping_data.angle >> 8;
      data[8] = ping_data.angle;

      sendPacket(9, DEBUGPKT, data);
      incPacketID();
   }
}
#endif

#ifdef NAVIGATE
int main( void ) {
   int16_t ping_distance;
   initAll();
   for(;;){
      ping_distance = ic_GetDistance();
      printf("%d, ", ping_distance);
      if(ping_distance < 12 && ping_distance > 0){
         printf("In if\n\r");
         SPI_Transfer(SEND_COMMAND);
         while(SPI_Transfer(NODATA) != SPI_ACK)
            printf("In while\n\r");
      }
   }
}
#endif

#ifdef NAVIGATE_AND_MAP
typedef enum {
   STOP_ROBOT,
   WAIT_SERVO,
   SWEEP_SERVO,
   WAIT_AFTER_SWEEP,
   DRIVING,
   TURNING,

} NavigateState; 

uint8_t data[9];
NavigateState state = STOP_ROBOT;

void RunNavigateSM( void ) {
   ping_vector ping_data;

   static uint8_t ping_angle;
   static uint16_t wait_time;

   switch(state) {
      case STOP_ROBOT:
         while(SPI_Transfer(SEND_COMMAND) != SPI_ACK); 
         SPI_Transfer(' ');//Stop the robot

         oc_TurnServo(ping_angle=0);  //Start the servo off at a 0 degree angle
         wait_time = 400;
         state = WAIT_SERVO;
         break;

      case WAIT_SERVO:
         _delay_ms(1);
         if(--wait_time == 0) {
            state = SWEEP_SERVO;
         }
         break;

      case SWEEP_SERVO:
         if(ping_angle%10 == 0) {
            doGetXY(data);    //Get our current x,y position
            ping_data = ic_PingSensorData();       //Gather the ping sensor data
            data[6] = ping_data.distance & 0xFF;   //Put the read distance into the data array
            data[7] = ping_data.angle >> 8;        //Put the read angle into the data array
            data[8] = ping_data.angle;
            sendPacket(9, DEBUGPKT, data);   //Send the packet to the observation station
            incPacketID();     //Increment the nonce
         }
         if(++ping_angle>180) {
            oc_TurnServo(90);
            wait_time = 400;
            state = WAIT_AFTER_SWEEP;
         } else {
            oc_TurnServo(ping_angle);  //Turn to the current i position
            wait_time = 10;
            state= WAIT_SERVO;
         }
         break;

      case WAIT_AFTER_SWEEP:
         _delay_ms(1);
         if(--wait_time == 0) {
            while(SPI_Transfer(SEND_COMMAND) != SPI_ACK);
            SPI_Transfer('w');
            wait_time =  40;
            state = DRIVING;
         }
         break;

      case DRIVING:
         doGetXY(data);
         if(--wait_time % 10==0) {
            ping_data = ic_PingSensorData(); //Gather the ping sensor data
            data[6] = ping_data.distance & 0xFF;   //Put the read distance into the data array
            data[7] = ping_data.angle >> 8;        //Put the read angle into the data array
            data[8] = ping_data.angle;
            sendPacket(9, DEBUGPKT, data);   //Send the packet to the observation station
            incPacketID();     //Increment the nonce
            if(ping_data.distance < 15) {
               while(SPI_Transfer(SEND_COMMAND) != SPI_ACK);
               SPI_Transfer('a');
               wait_time = 1000;
               state=TURNING;
            }
         }
         if(wait_time == 0) {
            while(SPI_Transfer(SEND_COMMAND) != SPI_ACK); 
            SPI_Transfer(' ');//Stop the robot
            state = STOP_ROBOT;
         }
         break;

      case TURNING:
         if(--wait_time % 40==0) {
            doGetXY(data);
            ping_data = ic_PingSensorData(); //Gather the ping sensor data
            data[6] = ping_data.distance & 0xFF;   //Put the read distance into the data array
            data[7] = ping_data.angle >> 8;        //Put the read angle into the data array
            data[8] = ping_data.angle;
            sendPacket(9, DEBUGPKT, data);   //Send the packet to the observation station
            incPacketID();     //Increment the nonce
         }
         if(wait_time == 0) {
            while(SPI_Transfer(SEND_COMMAND) != SPI_ACK); 
            SPI_Transfer(' ');//Stop the robot

            state = STOP_ROBOT;
         }
         break;
   }
}

int main( void ) {
   ping_vector ping_data;
   int suspended = 0;
   initAll();  //This will set the ping sensor to 90 degrees.
   SET_BIT(DDRE, DDE6);

   doGetXY(data);
   ping_data = ic_PingSensorData(); //Gather the ping sensor data
   data[6] = ping_data.distance & 0xFF;   //Put the read distance into the data array
   data[7] = ping_data.angle >> 8;        //Put the read angle into the data array
   data[8] = ping_data.angle;
   sendPacket(9, DEBUGPKT, data);   //Send the packet to the observation station
   incPacketID();     //Increment the nonce

   for(;;){
      gatherPacket();
      packetHandle();
      RunNWStateMachine();
      if(isLeader==1 && networkElements>1) {
         if(suspended == 1) {

            SET_BIT(PORTE, PE6);

            switch (state) {
               case DRIVING:
                  while(SPI_Transfer(SEND_COMMAND) != SPI_ACK);
                  SPI_Transfer('w');
                  break;
               case TURNING:
                  while(SPI_Transfer(SEND_COMMAND) != SPI_ACK);
                  SPI_Transfer('a');
                  break;
            }
            suspended = 0;
         }

         RunNavigateSM();

      } else { // Not Leader
         if ( suspended == 0 ) {

            CLR_BIT(PORTE, PE6);

            while(SPI_Transfer(SEND_COMMAND) != SPI_ACK); 
            SPI_Transfer(' ');//Stop the robot
            suspended = 1;
         }
      }
   }
}
#endif
