/*
   The Soma Project

   Copyright 2007 Rachel Pasetes, John Burr, Sean Eves Knudsen,
                  Thomas Gerdes, Andrew Parra, Erik Pasternak

  This file is part of The Soma Project.

  The Soma Project is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  The Soma Project is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with The Soma Project; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/

/*!@file
   @ingroup localize
   @brief iRobot Create control module

   This file contains code to control the iRobot Create over a UART
*/
#ifndef IROBOT_UART1
#define IROBOT_UART1
#endif
#include "irobot.h"
#include "usart.h"
#include <stdio.h>
#include "util.h"
#include <util/delay.h>


FILE* usart;

void iRobot_Init(void) {
//Depending on the define this chooses where to send iRobot commands.
#ifdef IROBOT_UART1 
	usart = fdevopen((void*)USART1_Transmit,(void*)USART1_Receive);
#else
	usart = fdevopen((void*)USART0_Transmit,(void*)USART0_Receive);
#endif
}

void iRobot_SongStore(int8_t slot, int8_t length, int8_t* song)
{
   int8_t i;
   fputc(140, usart);
   fputc(slot, usart);
   fputc(length, usart);
   for (i=0; i < 2*length; i++)
      fputc(song[i],usart);
}

void iRobot_SongPlay(int8_t slot)
{
   fputc(141, usart);
   fputc(slot, usart);
}

int iRobot_SongPlaying( void ) {
   fputc(142, usart);
   fputc(37, usart);
   return fgetc(usart);
}

///////////////////////////////////////////////
// activate_safe_mode()
// Activates safe mode on the iRobot.  Enables
// user control of the iRobot.  Turns off all
// LED's.
///////////////////////////////////////////////

void iRobot_ModeSafe(void)
{
   fputc(128, usart);
   fputc(131, usart);
}

///////////////////////////////////////////////
// activate_full_mode()
// Gives complete control of the iRobot.
// Turns off safety features of cliff sensors
// LED's.
///////////////////////////////////////////////
void iRobot_ModeFull(void)
{
   fputc(128, usart);
   fputc(132, usart);
   iRobot_PlayLedOn();
}

///////////////////////////////////////////////
// change_baud_irobot(uint8_t c)
// Changes baud rate on the iRobot.
// Baud Code: 	0		300
//				1		600
//				2		1200
//				3		2400
//				4		4800
//				5		9600
//				6		14400
//				7		19200
//				8		28800
//				9		38400
//				10		57600
//				11		115200
///////////////////////////////////////////////
void iRobot_ChangeBaud(uint8_t c)
{
	fputc(129, usart);
	fputc(c, usart);
}

void iRobot_Drive(int16_t left_speed, int16_t right_speed)
{
	uint8_t send_byte = right_speed>>8;
	fputc(145, usart);
	fputc(send_byte, usart);
	send_byte = right_speed & 0x00FF;
	fputc(send_byte, usart);
	send_byte = left_speed>>8;
	fputc(send_byte, usart);
	send_byte = left_speed & 0x00FF;
	fputc(send_byte, usart);
}

void iRobot_Drive_Distance_Fwd( int16_t distance, int16_t speed ){
   int16_t distance_traveled = 30;
   iRobot_GetDistance();
   while( distance >= distance_traveled ){
      distance_traveled += iRobot_GetDistance();
      printf("distance_traveled = %d\n\r", distance_traveled);
      iRobot_Drive(speed, speed);
   }
   iRobot_Drive(0, 0);
}

void iRobot_Drive_Distance_Rev( int16_t distance, int16_t speed ){
   int16_t distance_traveled = 30;
   iRobot_GetDistance();
   while( distance >= distance_traveled ){
      distance_traveled -= iRobot_GetDistance();
      printf("distance_traveled = %d\n\r", distance_traveled);
      iRobot_Drive(-speed, -speed);
   }
   iRobot_Drive(0, 0);
}

/*
void iRobot_Drive_Straight( int16_t angle ){
   if( angle != 0 )
}
*/

void iRobot_Wait(int8_t time){
	fputc(155, usart);
	fputc(time, usart);
}

int16_t iRobot_GetCliffLeft(void){
	fputc(142, usart);
	fputc(28, usart);
	return ((fgetc(usart)<<8) | (fgetc(usart)));
}

int16_t iRobot_GetCliffFrontLeft(void){
	fputc(142, usart);
	fputc(29, usart);
	return ((fgetc(usart)<<8) | (fgetc(usart)));
}

int16_t iRobot_GetCliffRight(void){
	fputc(142, usart);
	fputc(31, usart);
	return ((fgetc(usart)<<8) | (fgetc(usart)));
}

int16_t iRobot_GetCliffFrontRight(void){
	fputc(142, usart);
	fputc(30, usart);
	return ((fgetc(usart)<<8) | (fgetc(usart)));
}

int8_t iRobot_GetChargingState(void){
	fputc(142, usart);
	fputc(21, usart);
	return fgetc(usart);
}

int8_t iRobot_GetStallRight(void){
	fputc(142, usart);
	fputc(14, usart);
	return (fgetc(usart) & 0x08)>>3;
}

int8_t iRobot_GetStallLeft(void){
	fputc(142, usart);
	fputc(14, usart);
	return (fgetc(usart) & 0x10)>>4;
}

int8_t iRobot_GetPlayButton(void){
	fputc(142, usart);
	fputc(18, usart);
	return (fgetc(usart) & 0x01);
}

int8_t iRobot_GetAdvanceButton(void){
	fputc(142, usart);
	fputc(18, usart);
	return (fgetc(usart) & 0x04);
}

int16_t iRobot_GetDistance(void){
	fputc(142, usart);
	fputc(19, usart);
	return ((fgetc(usart)<<8) | (fgetc(usart)));
}

int16_t iRobot_GetAngle(void){
   int16_t angle;
   fputc(142, usart);
   fputc(20, usart);
   angle = (fgetc(usart)<<8) | (fgetc(usart));
   return(angle);
}

void iRobot_PlayLedOn(void){
   /*fputc(128, usart);
   fputc(132, usart);*/
   fputc(139, usart);
   fputc(2, usart);
   fputc(0, usart);
   fputc(0, usart);
}

void iRobot_PlayLedOff(void){
   fputc(139, usart);
   fputc(0, usart);
   fputc(0, usart);
   fputc(0, usart);
}
   

#ifdef MECH_DEBUG
/*char starWars[] = { 55, 32, 55, 32, 55, 32, 51,
                    16, 57, 16, 55, 32, 51, 16, 57, 16, 55,
                    32, 62, 32, 62, 32, 62, 32, 63, 16, 56,
                    16, 54, 32, 51, 16, 57, 16, 55, 32}; //36 bytes
*/
// C B C F --- D C D C B  B  - D C D F --- B A B A G
// 
// D 74
// C 72
// B 71
// A 69
// G 67
// F 65
// E 64
enum NOTES { 
   ND = 74, NC = 72, NB = 71, NA = 69, NG = 67, NF = 65, NE = 64, 
   NX = 30
};
enum DELAY {
   STH = 10, ETH = 20, QTR = 40, HLF = 80
};

char finalCountdown1[] = { 
   NC, STH,  NB, STH,  NC, ETH,  NF, QTR,    NX, QTR,
   ND, STH,  NC, STH,  ND, ETH,  NC, ETH,    NB, HLF, 
   NX, ETH,  ND, STH,  NC, STH,  ND, QTR,    NF, QTR,  NX, 3*ETH,
};
char finalCountdown2[] = { 
   NB, STH,   NA, STH,  NB, ETH,  NA, ETH,  NG, ETH,  NB, ETH,
   NA, 3*ETH, NC, STH,  NB, STH,  NC, QTR,  NF, HLF,  NX, STH,
   ND, STH,   NC, STH,  ND, ETH,  NC, ETH,  

};
char finalCountdown3[] = {
   NB, QTR,    NX, STH,  ND, STH,  NC, STH,  ND, QTR,  NF, QTR,
   NX, 3*ETH,  NB, STH,  NA, STH,  NB, ETH,  NA, ETH,  NG, ETH, NB, ETH,
   NA, 3*ETH,  NC, STH,  NB, STH,
};
char finalCountdown4[] = {
   NC, QTR,  NF, HLF,  NX, ETH,  ND, STH,  NC, STH, ND, ETH, NC, ETH,
   NB, HLF,  NX, ETH,  ND, STH,  NC, STH,  ND, QTR, NF, QTR, NX, 3*ETH,
   NB, STH,  NA, STH,
};
char finalCountdown5[] = {
   NB, ETH,  NA, ETH,  NG, ETH,  NB, ETH,  NA, 3*ETH, NC, STH, NB, STH,
   NC, QTR,  NF, HLF,  NX, STH,  ND, STH,  NC, STH,   ND, ETH, NC, ETH,
   NB, HLF,  NX, ETH,
};
char finalCountdown6[] = {
   ND, STH,  NC, STH,  ND, QTR,  NF, QTR,  NX, 3*ETH, NB, STH, NA, STH,
   NB, ETH,  NA, ETH,  NG, ETH,  NB, ETH,  NA, 3*ETH, NG, STH, NA, STH,
   NB, 3*ETH,
};
char finalCountdown7[] = {
   NA, STH,  NB, STH,  NC, ETH,  NB, ETH,  NA, ETH,  NG, ETH,  
   NF, QTR,  ND, QTR,  NX, STH/2, NC,  15 * STH/2, NC, 9*STH/2,
   ND, 7*STH/2, NC, STH, NB, STH, NC, 3*HLF,
};
int main(int argc, char**argv) {
   unsigned char i;
   asm("cli");
   USART0_Init( BAUD57600 );
   USART1_Init( BAUD57600 );
   asm("sei");
   printf("Hello World\n\r");
   iRobot_Init();
   iRobot_ModeFull();
   /*
   while(1){
      iRobot_Drive(200, 200);
      _delay_sec(2);
      printf("1.  get_distance = %d\n\r", iRobot_GetDistance());
      printf("1.1 get_angle = %d\n\r", iRobot_GetAngle());
      while(1);
      iRobot_Drive(-200, -200);
      _delay_sec(2);
      printf("2.  get_distance = %d\n\r", iRobot_GetDistance());
      printf("2.1 get_angle = %d\n\r", iRobot_GetAngle());
      iRobot_Drive(20, 20);
      _delay_sec(2);
      printf("3.  get_distance = %d\n\r", iRobot_GetDistance());
      printf("3.1 get_angle = %d\n\r", iRobot_GetAngle());
      iRobot_Drive(-20, -20);
      _delay_sec(2);
      printf("4.  get_distance = %d\n\r", iRobot_GetDistance());
      printf("4.1 get_angle = %d\n\r", iRobot_GetAngle());
      while(1);
   }
   while(1){
      printf("I am in the while loop\n\r");
      iRobot_Drive_Distance_Fwd(610, 200);
      printf("I just finished driving forward\n\r");
      _delay_sec(2);
      printf("I am now going to start driving backward\n\r");
      iRobot_Drive_Distance_Rev(610, 200);
      while(1);
   }
   */
   iRobot_ModeFull();
   _delay_ms(30);
   iRobot_PlayLedOn();
   _delay_ms(30);
   iRobot_SongStore(0, 16, finalCountdown1);
   iRobot_SongStore(1, 16, finalCountdown2);
   iRobot_SongStore(2, 16, finalCountdown3);
   iRobot_SongStore(3, 16, finalCountdown4);
   iRobot_SongStore(4, 16, finalCountdown5);
   iRobot_SongStore(5, 15, finalCountdown6);
   iRobot_SongStore(6, 15, finalCountdown7);
   iRobot_SongPlay(0);
   while(iRobot_SongPlaying()==1);

   iRobot_SongPlay(1);
   while(iRobot_SongPlaying()==1);

   iRobot_SongPlay(2);
   while(iRobot_SongPlaying()==1);

   iRobot_SongPlay(3);
   while(iRobot_SongPlaying()==1);

   iRobot_SongPlay(4);
   while(iRobot_SongPlaying()==1);
   iRobot_SongPlay(5);
   while(iRobot_SongPlaying()==1);
   iRobot_SongPlay(6);
   while(iRobot_SongPlaying()==1);
   for(;;) {}
   return 0;
}

#endif // ifdef MECH_DEBUG
