/*
   The Soma Project

  Copyright 2007 Rachel Pasetes, John Burr, Sean Eves Knudsen, 
                 Thomas Gerdes, Andrew Parra, Erik Pasternak

  This file is part of The Soma Project.

  The Soma Project is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  The Soma Project is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with The Soma Project; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/

/*!@defgroup observe Observation Station
   @brief Test application for drawing pixmaps in OpenGL.
   @file
   @brief Observation station main module
   @ingroup PixmapTest
*/

#include <GL/glut.h>
#include <stdio.h>
#include <stdlib.h>

#include "../command/Packet.h"
#include "../command/ic.h"
#include "mapData.h"

typedef struct {
   int16_t x;
   int16_t y;
   int16_t angle;
   ping_vector ping_data;
   char valid;
} robots;

int mouse1; /*!< State of left mouse button*/
int mouse2; /*!< State of right mouse button*/
int mouse3; /*!< State of middle mouse button*/
int lastx;  /*!< Last mouse X position */
int lasty;  /*!< Last mouse Y position */
int screenx, screeny;
int dx;     /*!< Render position X */
int dy;     /*!< Render position Y */
float scale;/*!< Render scale */

robots robo[4];
float color[4][4];

char *outputstrings[10];
void updateMap(int RoboID, ping_vector ping_data);

/*!@brief reposition Raster Position
   @param x X position on screen
   @param y Y position on screen

   This is a hack suggested in the Red Book (OpenGL) to set the
   RasterPos for anywhere off screen. This is needed to render pixmaps
   partially off screen.
   @note (0,0) is the top left of the screen. This corresponds to the
   mouse position on the screen.
*/
void setRasterPos(int x, int y) {
   glRasterPos2i(0,0);
   glBitmap (0, 0, 0, 0, x, -y, NULL);
}
void putstring(char *str) {
   int i;
   char *p = outputstrings[9];
   strncpy(p, str,100);
   for(i=9;i>0;i--) {
      outputstrings[i]=outputstrings[i-1];
   }
   outputstrings[0] = p;

}
void glprint( GLfloat x, GLfloat y, char *text ) {
   char *p;
   setRasterPos(x,y);
   for (p=text;*p;p++)
      glutBitmapCharacter(GLUT_BITMAP_HELVETICA_10, *p);
}

/*!@brief Render OpenGL
*/
void Display(void)
{
   unsigned char * map;
   int i, j;
   int bx,by;
   glClear(GL_COLOR_BUFFER_BIT);
   glPixelZoom(scale,-scale);
   glPixelStorei(GL_UNPACK_ALIGNMENT, 1); // Data is byte aligned

   for(i=-dx/(scale*MAPSIZE)-1;i<(screenx-dx)/(scale*MAPSIZE);i++) {
      for(j=dy/(scale*MAPSIZE)-1;j<(screeny+dy)/(scale*MAPSIZE);j++) {
         map = getMapData(i,j,0);
         if(map!=NULL) {
            setRasterPos(dx+i*scale*MAPSIZE, dy+-(j+1)*scale*MAPSIZE);
            glDrawPixels(MAPSIZE, MAPSIZE, GL_LUMINANCE, 
                         GL_UNSIGNED_BYTE, map);
         }
      }
   }
   glPushMatrix();
   glTranslated(dx, -dy, 0);
   glColor4f(1.0f,1.0f,1.0f, 0.5f);
   glBegin(GL_LINES);
      for(i=-dx/(scale*MAPSIZE); i<(screenx-dx)/(scale*MAPSIZE);i++) {
         glVertex2f(MAPSIZE*i*scale, dy); 
         glVertex2f(MAPSIZE*i*scale, screeny+dy); 
      }
      for(j=dy/(scale*MAPSIZE)-1;j<(screeny+dy)/(scale*MAPSIZE);j++) {
         glVertex2f(-dx, MAPSIZE*j*scale); 
         glVertex2f(screenx-dx, MAPSIZE*j*scale); 
      }
   glEnd();
   for(i=0;i<4;i++) {
      if(robo[i].valid) {
         glPushMatrix();
         glTranslatef(robo[i].x*scale, robo[i].y*scale, 0);
         glColor4fv(color[i]);
         glRectf(-2*scale, -2*scale, 2*scale, 2*scale);
         glBegin(GL_LINES);
            glVertex2f(0,0);
            float rad = 3.14 / 180 * (float) robo[i].angle;
            glVertex2f(20*scale*cos(rad), 20*scale*sin(rad));
            glVertex2f(0,0);
            rad = 3.14 / 180 * (float) robo[i].ping_data.angle;
            glVertex2f(robo[i].ping_data.distance*scale*cos(rad),
                       robo[i].ping_data.distance*scale*sin(rad));
         glEnd();
         glColor4f(1.0f,1.0f, 1.0f,1.0f);
         glPopMatrix();
      }
   }
   glPopMatrix();
   for(i=0;i<10;i++)
      glprint(0,-i*10, outputstrings[i]);
   glFlush();        //Finish rendering
   glutSwapBuffers();
}

/*!@brief Handle mouse clicks
   @param button Mouse button that was clicked
   @param state If the button is up or down
   @param x X position of the mouse on the screen
   @param y Y position of the mouse on the screen
*/
void mouseFunc(int button, int state, int x, int y) {
   ping_vector ping_data;
   char buf[100];
   float cx,cy;
   if(button==GLUT_LEFT_BUTTON) {
      if(state == GLUT_DOWN) {
         cx=(x-dx)/scale-robo[3].x;
         cy=(screeny-y+dy)/scale-robo[3].y;
         ping_data.distance = sqrt(cx*cx+cy*cy);
         if(x==0) {
            ping_data.angle = cy>0?90:270;
         } else {
            ping_data.angle = (atan(cy/cx) * 180 / 3.1415927);
            if(cx < 0 && cy >= 0) {
               ping_data.angle = 180 + ping_data.angle;
            } else if (cx < 0 && cy < 0) {
               ping_data.angle = ping_data.angle+180;
            } else if ( cy < 0 && cx >= 0) {
               ping_data.angle = 360 + ping_data.angle;
            }
            sprintf(buf, "x= %d y = %d a=%d; d=%d, a=%d",
                  robo[3].x, robo[3].y, robo[3].angle,
                  ping_data.distance, ping_data.angle);
         }
         putstring(buf);
         robo[3].ping_data.angle = ping_data.angle;
         robo[3].ping_data.distance = ping_data.distance;
         updateMap(3, ping_data);
      }
      //mouse1 = (state==GLUT_DOWN ? 1 : 0);
   }
   else if (button==GLUT_RIGHT_BUTTON) {
      if(state == GLUT_DOWN) {
         cx=(x-dx)/scale;
         cy=(screeny-y+dy)/scale;
         robo[3].x = cx;
         robo[3].y = cy;
      }
      //mouse2 = (state==GLUT_DOWN ? 1 : 0);
   }
   else if (button==GLUT_MIDDLE_BUTTON)
      mouse3 = (state==GLUT_DOWN ? 1 : 0);

   lastx = x;
   lasty = y;
}

/*!@brief Handle mouse movement
   @param x X position of the mouse on the screen
   @param y Y position of the mouse on the screen
*/
void motionFunc(int x, int y) {
   if(mouse3) {
      dx = dx - ( lastx-x );
      dy = dy - ( lasty-y );
      glutPostRedisplay();

   }
   lastx=x;
   lasty=y;
}

/*!@brief Handle the window being resized.
   @param x Width of the screen
   @param y Height of the screen
*/
void Reshape(int x, int y)
{
   if (y == 0 || x == 0) return;  //Nothing is visible then, so return
   screenx=x;screeny=y;
   //Set a new projection matrix
   glViewport(0,0,x,y);  //Use the whole window for rendering
   glMatrixMode(GL_PROJECTION);  
   glLoadIdentity();
   gluOrtho2D(0, 0+x, 0,y+0);
   //gluOrtho2D(0, x+0, y+0,0);
   glMatrixMode(GL_MODELVIEW);
   glLoadIdentity();
}

/*!@brief Handle normal key presses.
   @param key key pressed
   @param x X position of mouse on screen
   @param y Y position of mouse on screen
*/
void kbFunc(unsigned char key, int x, int y)
{
   if (key == '1' || key == 'w' || key == 's' || key == 'a' || key== 'd' || key ==' ') {
      sendPacket(1, 1, &key); //TODO: Change type
      incPacketID();
      printf("%c!\n", key);
   }
   if (key == 'b') {
	   sendPacket(1, START, &key);
      incPacketID();
   }
   if (key == 27) exit(0); // If the escape key is pressed, quit
}
/*!@brief Handle Up and Down keys.
   @param key key pressed
   @param x X position of mouse on screen
   @param y Y position of mouse on screen
*/
void specialkbFunc(int key, int x, int y) {
   switch(key) {
      case GLUT_KEY_UP:
         scale += 0.1;
         if(scale > 5.0) scale = 5.0;
         glutPostRedisplay();
         break;
      case GLUT_KEY_DOWN:
         scale -= 0.1;
         if(scale < 1.0) scale = 1.0;
         glutPostRedisplay();
         break;
   }
}

void updateMap(int RoboID, ping_vector ping_data) {
   unsigned char *map;
   int file_x,file_y;
   int map_x, map_y;
   int fx, fy;
   int dx, dy;
   int x, y;
   int ny;
   int angle;
   int tan_angle;
   int mappos;

   file_x =  robo[RoboID].x / (MAPSIZE);
   file_y =  robo[RoboID].y / (MAPSIZE);

   map_x= (robo[RoboID].x) % (int)(MAPSIZE);
   map_y= (robo[RoboID].y) % (int)(MAPSIZE);
   if(map_x < 0) {
      file_x--;      map_x += MAPSIZE;
   }
   if(map_y < 0) {
      file_y--;      map_y +=MAPSIZE;
   }
   map = getMapData(file_x,file_y,0);

   int *walk, *walkd, *walkf, *walkfi, *walkm;
   int *dep, *depd, *depfi, *depm;

   x=y=0;
   angle = (ping_data.angle + robo[RoboID].angle+270) % 360;

   if((angle>45 && angle<135) || (angle>225 && angle <315) ) {
      //Walk along Y
      tan_angle = (cosd(angle) << 15) / sind(angle); //really 1/tan
      fy = (ping_data.distance * sind(angle)) >> 15;
      fx = fy*tan_angle >> 15;

      walk  = &y;      dep   = &x;
      walkf = &fy;
      walkd = &dy;     depd  = &dx;
      walkfi= &file_y; depfi = &file_x;
      walkm = &map_y;  depm  = &map_x;

   } else {
      //Walk along X
      tan_angle = (sind(angle) << 15) / cosd(angle);
      fx = (ping_data.distance * cosd(angle)) >> 15;
      fy = fx*tan_angle >> 15;

      walk  = &x;      dep   = &y;
      walkf = &fx;
      walkd = &dx;     depd  = &dy;
      walkfi= &file_x; depfi = &file_y;
      walkm = &map_x;  depm  = &map_y;

   }
   dx=fx>x?1:-1;
   dy=fy>y?1:-1;

   while(*walk!=*walkf) {
      ny=((*walk) * tan_angle) >> 15;
      while((*dep)!=ny) {
         mappos = (MAPSIZE-(map_y)-1)*MAPSIZE+map_x;
         map[mappos] = 255;

         (*dep)+=*depd;
         (*depm)+=(*depd);
         if((*depm)<0 || (*depm)>=MAPSIZE) {
            saveMapData(file_x,file_y);
            (*depfi)+=(*depd);
            (*depm)=((*depm)+MAPSIZE)%MAPSIZE;
            map = getMapData(file_x,file_y,0);
         }
      }
      mappos = (MAPSIZE-(map_y)-1)*MAPSIZE+map_x;
      map[mappos] = 255;

      (*walk)+=(*walkd);
      (*walkm)+=(*walkd);
      if((*walkm)<0 || (*walkm)>MAPSIZE) {
         saveMapData(file_x,file_y);
         (*walkfi) +=(*walkd);
         (*walkm) = ((*walkm) + MAPSIZE)%MAPSIZE;
         map = getMapData(file_x,file_y,0);
      }
   }
   mappos = (MAPSIZE-(map_y)-1)*MAPSIZE+map_x;
   map[mappos] = 0;
}

/*!@brief Idle function. Do updates to map.
*/
void Idle(void) {
   unsigned char *map;
   int i;
   int map_x, map_y, x, y;

   ping_vector ping_data;

   packet pkt;

   if(hasPackets()){
      pkt = dequeuePacket();
      printf("%d, %d, %d\n", pkt.RoboID, pkt.type, pkt.length);
      if(pkt.type == DEBUGPKT) {
         robo[pkt.RoboID].valid = 1;
         robo[pkt.RoboID].x = (pkt.Data[0] << 8) | pkt.Data[1];
         robo[pkt.RoboID].y = (pkt.Data[2] << 8) | pkt.Data[3];
         robo[pkt.RoboID].angle = (pkt.Data[4] << 8) | pkt.Data[5];

         ping_data.distance = pkt.Data[6];
         ping_data.angle    = (pkt.Data[7]<<8) | pkt.Data[8];
         robo[pkt.RoboID].ping_data.angle = ping_data.angle;
         robo[pkt.RoboID].ping_data.distance = ping_data.distance;

         for(i=0;i<6;i+=2) { 
            printf("%d ",(int16_t)((pkt.Data[i]<<8) | pkt.Data[i+1]));
         }
         printf("%d ", ping_data.distance);
         printf("%d ", ping_data.angle);
         printf("\n");

         updateMap(pkt.RoboID, ping_data);
      } else {
         for(i=0;i<pkt.length;i++) { 
            printf("%02x ", pkt.Data[i]);
         }
         printf("'%s'\n", pkt.Data);
      }
	  printf("-----------------------------------------\n\r");
   }
   if(mouse1==1||mouse2==1) {
      map_x=(lastx-dx)/scale;
      map_y=(lasty-dy)/scale;
      x = map_x/(MAPSIZE);
      y = map_y/(MAPSIZE);
      map_x=map_x % (int)(MAPSIZE);
      map_y=map_y % (int)(MAPSIZE);
      if(map_x < 0) {
         map_x += MAPSIZE;
         x = x - 1;
      }
      if(map_y < 0) {
         map_y += MAPSIZE;
         y = y - 1;
      }

      map=getMapData(x,y,1);
      if(map) {
         if(mouse1 && map[map_y*MAPSIZE+map_x] < 240) {
            map[map_y*MAPSIZE+map_x]+=5;
         }
         else if(mouse2 && map[map_y*MAPSIZE+map_x] > 10) {
            map[map_y*MAPSIZE+map_x]-=5;
         }
         saveMapData(x,y);
      }
   }
   glutPostRedisplay();
}

void doexit() {
   printf("Exiting... \n");
}

/*!@brief Main program loop.

   Initialize glut, setup glutFuncs, and call the glut main loop.
   @note This function should never return
*/
int main(int argc, char **argv) {
   dx = 150;
   dy = -150;
   scale =1.0;

   char c;
   int counter;
   int i;
   for(i=0;i<10;i++) {
      outputstrings[i] = malloc(sizeof(char)*100);
      outputstrings[i][0]='\0';
   }
   for (i=0;i<4;i++)
      robo[i].valid = 0;
   color[1][0] = 1.0; color[1][1] = 0.0; color[1][2] = 0.0;   color[1][3]=.5;
   color[2][0] = 0.0; color[2][1] = 1.0; color[2][2] = 0.0;   color[2][3]=.5;
   color[3][0] = 0.0; color[3][1] = 0.0; color[3][2] = 1.0;   color[3][3]=.5;
 /* 
   robo[3].x=0;
   robo[3].y=0;
   robo[3].angle=90;
   robo[3].valid=1;
 */
   if(USART0_Init(5)==0) {
      if(Packet_Init()!=0)
         putstring("Could not communicate with RF module");
   } else {
      putstring("Could not open com port");
   }

   //Initialize GLUT
   glutInit(&argc, argv);
   glutInitDisplayMode(GLUT_DOUBLE | GLUT_RGB);
   glutInitWindowSize(300,300);

   glutCreateWindow("SOMA Observation station");
   glClearColor(0.0,0.0,0.0,0.0);

   glutDisplayFunc(Display);
   glutReshapeFunc(Reshape);
   glutKeyboardFunc(kbFunc);
   glutIdleFunc(Idle);
   glutSpecialFunc(specialkbFunc);
   glutMouseFunc(mouseFunc);
   glutMotionFunc(motionFunc);
   glEnable(GL_BLEND);
   glDepthMask(GL_FALSE);
   glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);

   //Let GLUT get the msgs
   atexit(doexit);
   glutMainLoop();
   return 0;
}

