/*
   The Soma Project

   Copyright 2007 Rachel Pasetes, John Burr, Sean Eves Knudsen,
                  Thomas Gerdes, Andrew Parra, Erik Pasternak

  This file is part of The Soma Project.

  The Soma Project is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  The Soma Project is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with The Soma Project; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/

/*!@file
   @ingroup localize
   @brief TOF Communication drivers

   This file contains code to communicate with the TOF module
   through USART0 running in MSPI mode.
*/

#include <avr/io.h>
#include <util/delay.h>
#include "TOFcom.h"
#include "usart.h"

void TOF_Init (void) 
{
   uint8_t i;
   for(i = 0; i < N_US; i++){
      TOF_Data.valid[i] = 0;
      TOF_Data.IR[i] = 0;
      TOF_Data.TOF[i] = 0;
   }
   USART0_Init(TOFBAUD);
   PORTE |= (1<<CS1) | (1<<CS2) | (1<<CS3);  // CS idles high
   DDRE |= (1<<CS1) | (1<<CS2) | (1<<CS3);   // CS lines driven
}

void TOF_Transfer(uint8_t* in_buf, uint8_t* out_buf, uint8_t len, uint8_t CS)
{
   int i;
   SELECT_TOF(CS);
   for(i = 0; i < len; i++)
   {
      in_buf[i] = USART0_Transfer(out_buf[i]);
   }
   _delay_us(50);    //Needs some delay to ensure all of a transfer is completed.
   DESELECT_TOF(CS);
}

uint8_t TOF_ExpectIR( uint8_t mcu ) {
   uint8_t reply;
   
   if ( (mcu < TOF1) || (mcu > TOF3) )
      return 0;
   
   _delay_us(50);
   if ( mcu == TOF1 ) {
      DESELECT_TOF(TOF2);
      DESELECT_TOF(TOF3);
   } else if ( mcu == TOF2 ) {
      DESELECT_TOF(TOF1);
      DESELECT_TOF(TOF3);
   } else if ( mcu == TOF3 ) {
      DESELECT_TOF(TOF1);
      DESELECT_TOF(TOF2);
   } else {
      return 0;
   }
   _delay_us(50);

   SELECT_TOF(mcu);
   while ( (reply = USART0_Transfer( EXPECTIR )) != EXPECTIR_ACK ) {
      //_delay_us(50);
      //DESELECT_TOF(TOF1);
      //_delay_us(50);
      //SELECT_TOF(TOF1);
      ;
   }
   _delay_us(50);
   DESELECT_TOF(mcu);
   return reply;
}

/*erik's
void TOF_ExpectIR (void)
{
   uint8_t out_cmd = EXPECTIR;
   uint8_t in_cmd;
   TOF_Transfer( (uint8_t*)&in_cmd, (uint8_t*)&out_cmd, TOF_CLEN, CS1);
   TOF_Transfer( (uint8_t*)&in_cmd, (uint8_t*)&out_cmd, TOF_CLEN, CS2);
   TOF_Transfer( (uint8_t*)&in_cmd, (uint8_t*)&out_cmd, TOF_CLEN, CS3);
}*/

uint8_t TOF_RequestData(uint8_t CS)
{
   uint8_t reply;
   uint8_t currentIndex;
   
   currentIndex = CS-CSOFFSET;

   SELECT_TOF(CS);
   do {
      reply = USART0_Transfer( NO_DATA );
   } while ( (reply != TOF_DATA) && (reply != TOF_INVALID) );
   _delay_us(50);
   DESELECT_TOF(CS);

   _delay_us(50);
   
   SELECT_TOF(CS);
   if ( reply == TOF_DATA ) {
      USART0_Transfer( TOF_DATA_ACK );
      TOF_Data.valid[currentIndex] = 1;
   } else {
      USART0_Transfer( TOF_INVALID_ACK );
      TOF_Data.valid[currentIndex] = 0;
   }
   _delay_us(50);
   DESELECT_TOF(CS);
   _delay_us(50);

   SELECT_TOF(CS);
   TOF_Data.IR[currentIndex] = USART0_Transfer( NO_DATA )<<8;
   TOF_Data.IR[currentIndex] |= USART0_Transfer( NO_DATA );
   TOF_Data.TOF[currentIndex] = USART0_Transfer( NO_DATA )<<8;
   TOF_Data.TOF[currentIndex] |= USART0_Transfer( NO_DATA );
   _delay_us(50);
   DESELECT_TOF(CS);

   /*if ( (TOF_Data.IR[currentIndex] < 800) || 
        (TOF_Data.IR[currentIndex] > 1200) || 
        (TOF_Data.TOF[currentIndex] < 600) || 
        (TOF_Data.TOF[currentIndex] > 18294) ) {
      TOF_Data.valid[currentIndex] = 0;
   }*/

   return reply;
}

/*uint8_t TOF_RequestData(uint8_t CS)
{
   uint8_t result = 0;
   uint8_t currentIndex;
   uint8_t out_cmd = TOF_DATA;
   uint8_t in_cmd;            //this byte is ignored.
   uint8_t out_buf[] = {0, 0, 0, 0, 0};
   uint8_t in_buf[TOF_DLEN];

   if(CS != CSALL) {
      currentIndex = CS-CSOFFSET;
      TOF_Transfer( &in_cmd, &out_cmd, TOF_CLEN, CS);
      // wait for TOF to ready data for transmit
      _delay_us(50);
      _delay_us(50);
      TOF_Transfer( in_buf, out_buf, TOF_DLEN, CS);
      if (in_buf[0] == TOF_DATA){
         TOF_Data.IR[currentIndex] = in_buf[IR_HIGH]<<8;
         TOF_Data.IR[currentIndex] |= in_buf[IR_LOW];
         TOF_Data.TOF[currentIndex] = in_buf[TOF_HIGH]<<8;
         TOF_Data.TOF[currentIndex] |= in_buf[TOF_LOW];
         TOF_Data.valid[currentIndex] = 1;
      } else if(in_buf[0] == NO_DATA) {
         TOF_Data.valid[currentIndex] = 0;
      } else {
         TOF_Data.valid[currentIndex] = 0;
         result = CS;
      }
   } else { //if CSALL
      for ( CS = CS1; CS <= CS3; CS++ ) {
         currentIndex = CS-CSOFFSET;
         TOF_Transfer( &in_cmd, &out_cmd, TOF_CLEN, CS);
         _delay_us(50);
         _delay_us(50);
         TOF_Transfer( in_buf, out_buf, TOF_DLEN, CS);
         if (in_buf[0] == TOF_DATA){
            TOF_Data.IR[currentIndex] = in_buf[IR_HIGH]<<8;
            TOF_Data.IR[currentIndex] |= in_buf[IR_LOW];
            TOF_Data.TOF[currentIndex] = in_buf[TOF_HIGH]<<8;
            TOF_Data.TOF[currentIndex] |= in_buf[TOF_LOW];
            TOF_Data.valid[currentIndex] = 1;
         } else if(in_buf[0] == NO_DATA) {
            TOF_Data.valid[currentIndex] = 0;
         } else {
            TOF_Data.valid[currentIndex] = 0;
            result = CS;
         }
      }
   }//else CSALL
   // returns zero or the last CSn on which TOF sent back neither TOF_DATA or NO_DATA
   return result;
}*/

      


