/*
   The Soma Project

   Copyright 2007 Rachel Pasetes, John Burr, Sean Eves Knudsen,
                  Thomas Gerdes, Andrew Parra, Erik Pasternak

  This file is part of The Soma Project.

  The Soma Project is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  The Soma Project is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with The Soma Project; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/

/*!@file
   @ingroup localize
   @brief TOF Communication drivers

   This file contains code to communicate with the TOF module
   through USART0 running in MSPI mode.
*/

#ifndef _TOFCOM_H_
#define _TOFCOM_H_

#include "../share/util.h"
#include "../share/SPIcommands.h"

// 399 = 20000 baud
// 3999 = 2000 baud
#define TOFBAUD 399 // Higher values are slower bauds

//Chip Select. Used for all functions that require a chip be chosen.
#define CS1 PE3
#define CS2 PE4
#define CS3 PE5
#define CSOFFSET 3   //CSn - CSOFFSET = index in TOF_Data

#define TOF1 CS1
#define TOF2 CS2
#define TOF3 CS3

#define TOF_CLEN  1  //command length
#define TOF_DLEN  5  //data length
#define IR_HIGH   1  //IR High Byte
#define IR_LOW    2
#define TOF_HIGH  3
#define TOF_LOW   4

#define CSALL  16     //Instructs Request Data to query all the Tinys

#define N_US   3     //number of US receivers.

typedef struct {
  uint8_t valid[3]; //0 for invalid, 1 for valid. positions 1,2,3 correspond to the atTinys
  uint16_t IR[3]; //ir on time for each attiny
  uint16_t TOF[3];  //tof for each attiny
} TOF_Data_t; 
//This stores the data returned by the RequestData function
TOF_Data_t TOF_Data; 

#define SELECT_TOF(pin) CLR_BIT( PORTE, pin )
#define DESELECT_TOF(pin) SET_BIT( PORTE, pin )

/*!@file
   @brief Initializes the TOF software

   Sets the initial baud rate for USART0 and initializes it to MSPI mode.
*/
void TOF_Init (void);

/*!@file
   @brief Transfers bytes on USART0.

   in_buf and out_buf must be the same length. This function transfers out len bytes in the
   out_buf and returns all the bytes read during transfer in the in_buf.
*/
void TOF_Transfer(uint8_t* in_buf, uint8_t* out_buf, uint8_t len, uint8_t CS);

/*!@file
   @brief Sends the EXPECTIR command to one of the three ATTinys.
*/
uint8_t TOF_ExpectIR( uint8_t mcu );

/*!@file
   @brief Requests data from each of the three ATTinys.

   This function requests IR On and TOF data from an ATtiny. Blocks until the ATtiny
   returns TOF_DATA or TOF_INVALID. Stores the data in TOF_Data, which should
   be accessed after calling this function.
*/
uint8_t TOF_RequestData(uint8_t CS);

#endif
