/*
   The Soma Project

   Copyright 2007 Rachel Pasetes, John Burr, Sean Eves Knudsen,
                  Thomas Gerdes, Andrew Parra, Erik Pasternak

  This file is part of The Soma Project.

  The Soma Project is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  The Soma Project is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with The Soma Project; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/

/*! @defgroup lib Shared Libraries
    @brief Shared libraries for use in both the uController and on the
    PC.
    @file mapData.c
    @brief Map Data Structure
    @ingroup lib
*/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "mapData.h"
#include "util.h"
#ifdef OBSERVE
#define MAXLOADED 128 //shit, can't go over 128?
#else
#define MAXLOADED 10
#endif

/*! \brief 512 byte structure to hold map data. */
typedef struct MapDataStruct {
   uint16_t MapNum; /*!< UID of map this file belongs to */
   int16_t x;               /*!< X coord of top left corner */
   int16_t y;               /*!< Y coord of top left corner */
   int8_t filler[512-MAPSIZE*MAPSIZE-sizeof(int16_t)*3];
                            /*!< Whitespace to fill up 512 bytes */
   uint8_t map[MAPSIZE][MAPSIZE]; /*!< Map pixmap */
} MapData;

/*! Status of MapCache data blocks */
typedef enum MapCacheStatus {
   INVALID=0,
   VALID=1, /*!< There is a map loaded in this Cache block */
   DIRTY=2, /*!< If dirty is set, loaded map has not been written */
   LOCKED=4 /*!< Do not remove this block from memory while modifying */
} MapCacheStatus;

/*!\brief Cache for map files.
 *
 * Store 8 map data files in memory. This data
 * structure keeps track of which files are in memory as well as which
 * need to be written to disk. */
struct MapCache {
   MapCacheStatus status; /*!< validity of this entry */
   char filename[13];     /*!< File name of file loaded in this block */
   MapData data;          /*!< Map File Data */
} loadedMaps[MAXLOADED];

static int LRU; /*!< Least Recently used element of loadedMaps */

/*! \brief Create New File for map data
    \param x x coord of map
    \param y y coord of map
    \param index Index of loadedMaps array to load data into
    \return Pointer to loadedMaps[index].map (map pixmap)
*/
unsigned char *newMapData(int x, int y, int index, char lock) {
   int i,j;
   FILE *output;
   MapData *map;

   map = &(loadedMaps[index].data);

   map->MapNum = 1;
   map->x = x;
   map->y = y;

   for(i=0;i<59;i++) map->filler[i]=0;
   for(i=0;i<MAPSIZE;i++)
      for(j=0;j<MAPSIZE;j++)
         map->map[i][j]=0x80;

   output = fopen(loadedMaps[index].filename, "w");
   if(output != NULL) {
      fwrite(map, 1, 512, output);
      fclose(output);
      loadedMaps[index].status=VALID;
      return (unsigned char *)map->map;
   }

   return 0;
}

/*! \brief Attempt to load file with map data

    If there is no map data in the file to load, this function calls
    newMapData to create the map file.
    \param x x coord of map
    \param y y coord of map
    \param i Index of loadedMaps array to load data into
    \return Pointer to loadedMaps[i].map (map pixmap)
*/
unsigned char *loadMapData(int x, int y, int i, char lock) {
   FILE *input=NULL;
   MapData *map;
   char filename[13]="00000000.map";
   shortToHex(filename  , x);
   shortToHex(filename+4, y);

   map = &(loadedMaps[i].data);
   //If the map is currently valid, but needs to be written
   if((loadedMaps[i].status & VALID)) {
      if(loadedMaps[i].status & DIRTY) {
         input = fopen(loadedMaps[i].filename, "w");
         if(input != NULL) {
            fwrite(map, 1, 512, input);
            fclose(input);
         }
      }
   }
   strncpy(loadedMaps[i].filename, filename, 13);
   input = fopen(filename, "r");
   if(input!=NULL) {
      fread(map, 1, 512, input);
      fclose(input);
      loadedMaps[i].status = VALID;
      if(lock) loadedMaps[i].status |= LOCKED;
      return (unsigned char *)map->map;
   }

   return newMapData(x, y, i, lock);
}

unsigned char *getMapData(int x, int y, char lock) {
   int i;
   //Scan all the loaded, Valid maps
   for(i=0;i<MAXLOADED;i++) {
      if((loadedMaps[i].status & VALID)==VALID) {
         //If this map is the requested map, return it
         if(loadedMaps[i].data.x==x && loadedMaps[i].data.y==y) {
            return (unsigned char *)loadedMaps[i].data.map;
         }
      } else {
         //If there are invalid maps, fill that cache block
         return loadMapData(x,y,i,lock);
      }
   }
   do {
      LRU=(LRU+1)%MAXLOADED;
   } while(loadedMaps[LRU].status & LOCKED);
   return loadMapData(x,y,LRU,lock);
}


void saveMapData(int x, int y) {
   int i;
   for(i=0;i<MAXLOADED;i++) {
      if(loadedMaps[i].data.x==x && loadedMaps[i].data.y==y) {
         loadedMaps[i].status |= DIRTY;
         loadedMaps[i].status &= ~LOCKED;
         //printf("%d\n", loadedMaps[i].status);
         return;
      }
   }
}
