/*
   The Soma Project

   Copyright 2007 Rachel Pasetes, John Burr, Sean Eves Knudsen,
                  Thomas Gerdes, Andrew Parra, Erik Pasternak

  This file is part of The Soma Project.

  The Soma Project is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  The Soma Project is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with The Soma Project; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/

#ifndef _USART_H_
#define _USART_H_
/**@file
   @brief Include file for USART functions.
   @sa usart
   @ingroup usart

   @defgroup usart USART driver
   @code   #include "usart.h" @endcode

   This module has the functions related to the USARTs on the AVRs.
   This module was designed to be used universally on several processors
   (mega88, mega128, mega1281), and multiple targets (command,
   localize), several compile time switches must be set in order to
   chose which USART to enable and in what mode to initilize it in.

   @{
   @def USART0_UART
    @brief Enable USART0 as a UART
   @def USART0_SPI
    @brief Enables USART0 as a SPI
   @def USART1_UART
    @brief Enable USART1 as a UART
   @def USART1_SPI
    @brief Enables USART1 as a SPI
*/
#if defined(__DOXYGEN__)
#define USART0_UART
//#define USART1_UART
#define USART0_SPI
//#define USART1_SPI
#endif
/*!@enum BaudRate Baudrates for the USARTs.
   @note These baud rates assume a 16MHz clock.  These values need to be
   adjusted if the clock is changed.
*/
enum BaudRate {
   BAUD2400 =416,///<   2400 Baud with  0.1% Error
   BAUD4800 =207,///<   4800 Baud with  0.2% Error
   BAUD9600 =103,///<   9600 Baud with  0.2% Error
   BAUD14400 =68,///<  14400 Baud with  0.6% Error
   BAUD19200 =51,///<  19200 Baud with  0.2% Error
   BAUD28800 =34,///<  28800 Baud with -0.8% Error
   BAUD38400 =25,///<  38400 Baud with  0.2% Error
   BAUD57600 =16,///<  57600 Baud with  2.1% Error
   BAUD76800 =12,///<  76800 Baud with  0.2% Error
   BAUD115200 =8,///< 115200 Baud with -3.5% Error
   BAUD230400 =3,///< 230400 Baud with  8.5% Error
   BAUD250000 =3,///< 250000 Baud with  0.0% Error
   BAUD500000 =1,///< 500000 Baud with  0.0% Error
   BAUD1000000=0 ///<1000000 Baud with  0.0% Error
};

#ifdef USART0_UART
/*!@param baudrate Baudrate to initialize to. See BaudXXXX defines.
   @brief All of the initilization for UART communication on USART0

   This function sets the baudrate of the USART and sets up
   communication options.  It also enables the data received interrupt.
   @note Interrupts must be enabled after this function is called.
*/
inline void USART0_Init( enum BaudRate baudrate );

/*!@brief Sets the baud rate for USART0
*/
void USART0_SetBaud( enum BaudRate baudrate );

/*!@brief Recieve a byte from USART0.

   This function pulls a byte off the USART0 recieve buffer.  If the
   buffer is empty, it waits until it gets data in.
*/
unsigned char USART0_Receive( void );

/*!@brief Send data over USART0.

   This function puts a byte on the transmit buffer and then enables the
   transmit interrupt. The data is not sent immediately, but when the
   ISR handles it.
*/
void USART0_Transmit( unsigned char data );

/*!@brief Check input buffer for data.
   @returns 0 if there is no data in the queue, 1 if there is data.
*/
unsigned char USART0_HasData( void );

/*!@brief See if the send buffer has data in it.

   This function checks if the transmit buffer has data in it.
*/
unsigned char USART0_SendingData( void );

/*!@brief flushes the transmit buffer for USART0.
*/
void USART0_Flush( void );

#endif // ifdef USART0_UART


#ifdef USART0_SPI
/*!@brief Initializes USART0 in SPI mode.
   
   This function initializes the USART0 in SPI mode, taking
   in a defined baud rate. It sets the data mode to 0 (idle low,
   valid on rising edge).
*/
inline void USART0_Init( enum BaudRate baudrate );

/*!@brief transfers a byte of data over USART0 as SPI.
*/
unsigned char USART0_Transfer( unsigned char data );

/*!@brief Sets the baud rate for USART0
*/
void USART0_SetBaud( enum BaudRate baudrate );

#endif //USART0_SPI

#ifdef USART1_UART
/*!@param baudrate Baudrate to initialize to. See BaudXXXX defines.
   @brief All of the initilization for UART communication on USART1

   This function sets the baudrate of the USART and sets up
   communication options.  It also enables the data received interrupt.
   @note Interrupts must be enabled after this function is called.
*/
inline void USART1_Init( enum BaudRate baudrate );

/*!@brief Sets the baud rate for USART1
*/
void USART1_SetBaud( enum BaudRate baudrate );

/*!@brief Check input buffer for data.
   @returns 0 if there is no data in the queue, 1 if there is data.
*/
unsigned char USART1_HasData( void );

/*!@brief flushes the transmit buffer for USART1.
*/
void USART1_Flush( void );

/*!@brief Recieve a byte from USART1.

   This function pulls a byte off the USART1 recieve buffer.  If the
   buffer is empty, it waits until it gets data in.
*/
unsigned char USART1_Receive( void );

/*!@brief Send data over USART1.

   This function puts a byte on the transmit buffer and then enables the
   transmit interrupt. The data is not sent immediately, but when the
   ISR handles it.
*/
void USART1_Transmit( unsigned char data );

/*!@brief See if the send buffer has data in it.

   This function checks if the transmit buffer has data in it.
*/
unsigned char USART1_SendingData( void );

#endif // ifdef USART1_UART
#ifdef USART1_SPI
/*!@brief Initializes USART1 in SPI mode.
   
   This function initializes the USART1 in SPI mode, taking
   in a defined baud rate. It sets the data mode to 0 (idle low,
   valid on rising edge).
*/
inline void USART1_Init( enum BaudRate baudrate );

/*!@brief transfers a byte of data over USART1 as SPI.
*/
unsigned char USART1_Transfer( unsigned char data );

#endif //USART0_SPI
/**@}*/

#endif //ifndef _USART_H_
