/*
   The Soma Project

   Copyright 2007 Rachel Pasetes, John Burr, Sean Eves Knudsen,
                  Thomas Gerdes, Andrew Parra, Erik Pasternak

  This file is part of The Soma Project.

  The Soma Project is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  The Soma Project is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with The Soma Project; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/

/*!@file
   @brief Interrupt Driven UART0 driver.

      This is a simple bidirectional blocking UART driver.
   @sa usart
   @ingroup usart
*/
#include "usart.h"
#include <avr/io.h>
#include <avr/interrupt.h>
#include <stdio.h>
#include <util/delay.h>

#if defined(USART0_UART) && defined(USART0_SPI)
   #error USART0 initialized in both UART and SPI modes
#elif !defined(USART0_UART) && !defined(USART0_SPI)
	#error USART0 not initialized, but attempting to compile it
#endif

/* Hack USART0 vect for atmega88 */
#ifdef _AVR_IOMX8_H_
   #undef USART0_RX_vect
   #undef USART0_UDRE_vect
   #define USART0_RX_vect   USART_RX_vect
   #define USART0_UDRE_vect USART_UDRE_vect
#endif

/* UART Buffer Defines */
#define USART0_RX_BUFFER_SIZE 128 ///< Size of USART0 recieve buffer
#define USART0_TX_BUFFER_SIZE 128 ///< Size of USART0 transmit buffer
#define USART0_RX_BUFFER_MASK ( USART0_RX_BUFFER_SIZE - 1 )
///< Mask transmit buffer size for use in circular buffer
#define USART0_TX_BUFFER_MASK ( USART0_TX_BUFFER_SIZE - 1 )
///< Mask recieve buffer size for use in circular buffer
#if ( USART0_RX_BUFFER_SIZE & USART0_RX_BUFFER_MASK )
   #error RX buffer size is not a power of 2
#endif
#if ( USART0_TX_BUFFER_SIZE & USART0_TX_BUFFER_MASK )
   #error TX buffer size is not a power of 2
#endif
#ifdef USART0_SPI
#define XCK0_DDR DDRE
#define XCK0 PE2
#endif

#ifdef USART0_UART

/* Static Variables */
static unsigned char USART0_RxBuf[USART0_RX_BUFFER_SIZE];
static volatile unsigned char USART0_RxHead;
static volatile unsigned char USART0_RxTail;
static unsigned char USART0_TxBuf[USART0_TX_BUFFER_SIZE];
static volatile unsigned char USART0_TxHead;
static volatile unsigned char USART0_TxTail;
static volatile char status;

/*!@brief Interrupt service routine for USART0 recieve data.
*/
ISR(USART0_RX_vect) {
   unsigned char data;
   unsigned char tmphead;
   data = UDR0;
   tmphead = ( USART0_RxHead + 1 ) & USART0_RX_BUFFER_MASK;
   USART0_RxHead = tmphead;
   USART0_RxBuf[tmphead] = data; /* Store received data in buffer */
}

/*!@brief Interrupt service routine for USART0 transmit data.
*/
ISR(USART0_UDRE_vect) {
   unsigned char tmptail;

   if ( USART0_TxHead != USART0_TxTail ) {
//      if (PIND & _BV(PD2) ) {
//         UCSR0B &= ~(_BV(UDRIE0));
//         EIMSK |= _BV(INT0);
//      } else {
         tmptail = ( USART0_TxTail + 1 ) & USART0_TX_BUFFER_MASK;
         USART0_TxTail = tmptail;
         UDR0 = USART0_TxBuf[tmptail]; /* Start transmition */
//      }
   } else {
      UCSR0B &= ~(1<<UDRIE0);      /* No data, disable transmission */
   }
}
#endif

ISR(INT0_vect) {
   EIMSK &= ~(1<<INT0);
   UCSR0B |= (1<<UDRIE0);  /* Enable transmit interrupt */
}

#ifdef USART0_UART
inline void USART0_Init( enum BaudRate baudrate ) {
   USART0_SetBaud(baudrate);      /* Set the baud rate */
   #ifdef _AVR_IOMX8_H_    /* If is atmega88 */
   UCSR0A |= (1 << U2X0);  /* Enable 2x transmission if on an ATMega88*/

   //PIND |= _BV(PD2);
   //DDRD &= ~_BV(PD2);
   //EIMSK &= ~_BV(INT0);

   #else  /* atmega1281 */
   UCSR0A &= ~(1 << U2X0);
   #endif

   /* Set frame format: 8 data 1stop */
   UCSR0C  = (1<<UCSZ01) | (1<<UCSZ00);


   /* Enable UART receiver and transmitter */
   UCSR0B = ( ( 1 << RXCIE0 ) | ( 1 << RXEN0 ) | ( 1 << TXEN0 ) );
   /* Flush receive buffer */
   USART0_RxTail = USART0_RxHead = 0;
   USART0_TxTail = USART0_TxHead = 0;
   if(status==0) { // Initialize stdio only once
      stdout = stdin = fdevopen((void*)USART0_Transmit,
                                (void*)USART0_Receive);
      status=1;
   }
}
#endif
#ifdef USART0_SPI 
inline void USART0_Init(enum BaudRate baudrate) {
   UBRR0 = 0;
   /*Set XCK0 to output*/
   XCK0_DDR |= (1<<XCK0);
   /*Set the USART to MSPIM and the data mode to 0*/
   UCSR0C = (1<<UMSEL01) | (1<<UMSEL00);
   //Enable USART TxD and RxD
   UCSR0B = (1<<RXEN0) | (1<<TXEN0);
   //in MSPI mode UBRR must be set AFTER the outputs are enabled.
   USART0_SetBaud(baudrate);
}

unsigned char USART0_Transfer( unsigned char data )
{
   while (!(UCSR0A & (1<<UDRE0)));
   UDR0 = data;
   while (!(UCSR0A & (1<<RXC0))); 
   return UDR0;
}

#else
unsigned char USART0_Receive( void )
{
   unsigned char tmptail;
   while ( USART0_RxHead == USART0_RxTail ); /* Wait for incomming data */
   tmptail = ( USART0_RxTail + 1 ) & USART0_RX_BUFFER_MASK;
   USART0_RxTail = tmptail;
   return USART0_RxBuf[tmptail];
}

void USART0_Transmit( unsigned char data )
{
   unsigned char tmphead;
   tmphead = ( USART0_TxHead + 1 ) & USART0_TX_BUFFER_MASK; 
   while ( tmphead == USART0_TxTail ); /* wait while buffer full */
   USART0_TxBuf[tmphead] = data;
   USART0_TxHead = tmphead;
   UCSR0B |= (1<<UDRIE0);  /* Enable transmit interrupt */
}


unsigned char USART0_HasData( void )
{
   return ( USART0_RxHead != USART0_RxTail ); 
}

void USART0_Flush( void )
{
   asm("cli");
   USART0_TxHead = USART0_TxTail;  //The queue is circular so you
   asm("sei");
}                                   //can set tail to head, as well.


unsigned char USART0_SendingData( void )
{
   return ( USART0_TxHead != USART0_TxTail);
}
#endif


void USART0_SetBaud( enum BaudRate baudrate )
{
   int i = SREG & (_BV(SREG_I)); //gets the current status of global interrupts
   if(i) asm("cli");
   UBRR0H = (unsigned char) (baudrate>>8);
   UBRR0L = (unsigned char) baudrate;
   if(i) asm("sei");
}




#ifdef DEBUG_UART
/*!@brief Testbench for USART code
*/
void main( void )
{
   USART0_Init( BAUD57600 );

   asm("sei"); /* Enable interrupts => enable UART interrupts */

   for(;;) {
      USART0_Transmit( USART0_Receive() ); /* Echo the received character */
   }
}
#endif

#ifdef DEBUG_MSPI
/*!@brief Testbench for USART in MSPI mode*/
void main (void)
{
   int i;
   USART0_Init(BAUD500000);
   asm("sei");
   for(;;)
      USART0_Transfer(i++);
}
#endif
