/*
   The Soma Project

   Copyright 2007 Rachel Pasetes, John Burr, Sean Eves Knudsen,
                  Thomas Gerdes, Andrew Parra, Erik Pasternak

  This file is part of The Soma Project.

  The Soma Project is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  The Soma Project is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with The Soma Project; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/

/*!@file
   @brief Interrupt Driven UART0 driver.

      This is a simple bidirectional blocking UART driver.
   @sa usart
   @ingroup usart

 */
#include "usart.h"
#include <avr/io.h>
#include <avr/interrupt.h>
#include <stdio.h>

#if defined(USART1_UART) && defined(USART1_SPI)
   #error USART1 initialized in both UART and SPI modes
#elif !defined(USART1_UART) && !defined(USART1_SPI)
	#error USART1 not initialized, but attempting to compile it
#endif

/* UART Buffer Defines */

#define USART1_RX_BUFFER_SIZE 128 ///< Size of USART1 recieve buffer
#define USART1_TX_BUFFER_SIZE 128 ///< Size of USART1 transmit buffer
#define USART1_RX_BUFFER_MASK ( USART1_RX_BUFFER_SIZE - 1 )
///< Mask buffer size for use in circular buffer
#define USART1_TX_BUFFER_MASK ( USART1_TX_BUFFER_SIZE - 1 )
///< Mask buffer size for use in circular buffer
#if ( USART1_RX_BUFFER_SIZE & USART1_RX_BUFFER_MASK )
   #error RX buffer size is not a power of 2
#endif
#if ( USART1_TX_BUFFER_SIZE & USART1_TX_BUFFER_MASK )
   #error TX buffer size is not a power of 2
#endif

#ifdef USART1_UART

/* Static Variables */
static unsigned char USART1_RxBuf[USART1_RX_BUFFER_SIZE];
static volatile unsigned char USART1_RxHead;
static volatile unsigned char USART1_RxTail;
static unsigned char USART1_TxBuf[USART1_TX_BUFFER_SIZE];
static volatile unsigned char USART1_TxHead;
static volatile unsigned char USART1_TxTail;
static volatile char status;


/*!@brief Interrupt service routine for USART1 recieve data.
*/
ISR(USART1_RX_vect)
{
   unsigned char data;
   unsigned char tmphead;
   data = UDR1;
   tmphead = ( USART1_RxHead + 1 ) & USART1_RX_BUFFER_MASK;
   USART1_RxHead = tmphead;
   USART1_RxBuf[tmphead] = data; /* Store received data in buffer */
}

/*!@brief Interrupt service routine for USART1 transmit data.
*/
ISR(USART1_UDRE_vect)
{
   unsigned char tmptail;
   if ( USART1_TxHead != USART1_TxTail ) {
      tmptail = ( USART1_TxTail + 1 ) & USART1_TX_BUFFER_MASK;
      USART1_TxTail = tmptail;
      UDR1 = USART1_TxBuf[tmptail]; /* Start transmition */
   } else {
      UCSR1B &= ~(1<<UDRIE1);      /* No data, disable transmission */
   }
}
#endif

inline void USART1_Init( enum BaudRate baudrate )
{

   /* Enable 2x transmission */
   //UCSR1A |= (1 << U2X1);
   /* Enable UART receiver and transmitter */

   #ifdef USART1_SPI /*Sets it up as an SPI Master if defined*/
   /*Set XCK0 to output*/
   XCK0_DDR |= (1<<XCK0);
   /*Set the USART to MSPIM and the data mode to 0*/
   UCSR0C = (1<<UMSEL01) | (1<<UMSEL00) & ~(0<<UCPHA0) & ~(0<<UCPOL0);
   #else /*Or as a UART if not defined*/
   /* Set frame format: 8 data 1stop */
   UCSR1C  = (1<<UCSZ11) | (1<<UCSZ10);
   #endif

   UCSR1B = ( ( 1 << RXCIE1 ) | ( 1 << RXEN1 ) | ( 1 << TXEN1 ) );
   /* Set frame format: 8 data 1stop */
   /* Set the baud rate */
   USART1_SetBaud(baudrate);
   /* Flush receive buffer */
   USART1_RxTail = USART1_RxHead = 0;
   USART1_TxTail = USART1_TxHead = 0;
//   if(status==0) { // Initialize stdio only once
//      stdout = stdin = fdevopen((void*)USART1_Transmit,
//                                (void*)USART1_Receive);
//      status=1;
//   }
}

void USART1_SetBaud( enum BaudRate baudrate )
{
   int i = SREG & (_BV(SREG_I)); //gets the current status of global interrupts
   if(i) asm("cli");
   UBRR1H = (unsigned char) (baudrate>>8);
   UBRR1L = (unsigned char) baudrate;
   if(i) asm("sei");
}

#ifdef USART1_SPI
unsigned char USART1_Transfer( unsigned char data )
{
   /*Wait for the transmit buffer to be clear*/
   while (!(UCSR0A & (1<<UDRE0)));
   UDR0 = data;
   /*Wait for valid data to be ready*/
   while (!(UCSR0A & (1<<RXC0)));
   return UDR0;
}

#else
unsigned char USART1_Receive( void )
{
   unsigned char tmptail;
   while ( USART1_RxHead == USART1_RxTail ); /* Wait for incomming data */
   tmptail = ( USART1_RxTail + 1 ) & USART1_RX_BUFFER_MASK;
   USART1_RxTail = tmptail;
   return USART1_RxBuf[tmptail];
}

void USART1_Transmit( unsigned char data )
{
   unsigned char tmphead;
   tmphead = ( USART1_TxHead + 1 ) & USART1_TX_BUFFER_MASK; 
   while ( tmphead == USART1_TxTail ); /* wait while buffer full */
   USART1_TxBuf[tmphead] = data;
   USART1_TxHead = tmphead;
   UCSR1B |= (1<<UDRIE1);  /* Enable transmit interrupt */
}
#endif

unsigned char USART1_HasData( void )
{
   return ( USART1_RxHead != USART1_RxTail ); 
}

void USART1_Flush( void )
{
   USART1_TxHead = USART1_TxTail;  //The queue is circular so you
}                                   //can set tail to head, as well.

unsigned char USART1_SendingData( void )
{
   return ( USART1_TxHead != USART1_TxTail);
}

#ifdef DEBUG_UART
/*!@fn void main( void )
   @brief Testbench for USART code
*/
void main( void )
{
   char c=0;
   asm("cli");
   initRf();
   USART0_Init( BAUD57600 );   /* Set the baudrate to 57.6 kbps @ 8 MHz clock */
   //USART1_Init( BAUD57600 );
   asm("sei");

   for(;;) {
      if(USART1_HasData()) {
         c=USART1_Receive();
         USART1_Transmit(c);
         USART0_Transmit(c);
      }
      if(USART0_HasData()) {
         c=USART0_Receive();
         USART0_Transmit(c);
         USART1_Transmit(c);
      }
   }
   return 0;
}
#endif
