/*
   The Soma Project

   Copyright 2007 Rachel Pasetes, John Burr, Sean Eves Knudsen,
                  Thomas Gerdes, Andrew Parra, Erik Pasternak

  This file is part of The Soma Project.

  The Soma Project is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  The Soma Project is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with The Soma Project; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/

/*!@file tcnt1.c
   @brief Timer/Couner1 Input Capture driver
   @ingroup tof
*/

#include <avr/io.h>
#include <avr/interrupt.h>
//#include <assert.h>
//#include <stdlib.h>
#include "../share/util.h"

//#define TCNT1_TEST_IC 1

volatile uint8_t TCNT1_EdgeTypeToggle;
volatile uint16_t TCNT1_ICTimerValue;
volatile uint8_t TCNT1_ICOccurred;

#define TCNT1_DIR_REG DDRA
#define TCNT1_ICP_REG PORTA
#define TCNT1_ICP_PIN PA7

/*!@brief Initialize Timer/Counter1 for input capture use.
   
   Sets all necessary port directions and configures Timer/Counter1 for 
   input capture operation.

   @param edgeType Sets initial edge type for trigger.  
                   0 for negative edge.  1 for positive edge.
   @param edgeType Toggle Sets whether to toggle edge select in ISR.  
                   0 for no change, 1 to toggle.
   @note Interrupts must be enabled after this function is called.
 */
void TCNT1_InitIC( uint8_t edgeType, uint8_t edgeTypeToggle )
{
   // set ICP as Hi-Z input
   CLR_BIT( TCNT1_DIR_REG, TCNT1_ICP_PIN );
   CLR_BIT( TCNT1_ICP_REG, TCNT1_ICP_PIN );

   // set Waveform Generation Mode to Normal (use all 16 bits)
   CLR_BIT( TCCR1A, WGM11 );
   CLR_BIT( TCCR1A, WGM10 );   

   // Trigger on edgeType, WGM is Normal, clock/8 from prescaler
   TCCR1B = (edgeType<<ICES1) | (1<<CS11);
   
   // Enable input capture interrupt
   SET_BIT( TIMSK1, ICIE1 );
   SET_BIT( TIFR1, ICF1 );

   TCNT1_ICOccurred = 0;
   TCNT1_EdgeTypeToggle = edgeTypeToggle;

   #ifdef TCNT1_TEST_IC
   // set PA0 as output low
   SET_BIT( DDRA, PA0 );
   CLR_BIT( PORTA, PA0 );
   #endif
}

/*!@brief Enables the input capture interrupt.
 *
 * @note Momentarily disables global interrupts and restores I-bit.
 */
inline void TCNT1_EnableIC( void )
{
   uint8_t tempSREG = SREG;
   asm("cli");
   SET_BIT( TIMSK1, ICIE1 );
   SET_BIT( TIFR1, ICF1 );
   SREG = tempSREG;
}

/*!@brief Blocking function to wait for input capture ISR to set
 *        TCNT1_ICOccurred flag.
 *
 * Introduces a 1 tick error if TCNT1_ICTimerValue lands on 0, this
 * allows 0 to be a timeout error. Uses the OC to generate a timeout.
 * @return Timer value when most recent input capture even occurred.
 */
/*uint16_t TCNT1_WaitForICOccurred( void )
{
   uint16_t result;
   OCR1A = TCNT1 - 1;    // timeout at 1 less than the current timer value
   TIFR1 = _BV(OCF1A);  // clear OC flag
   while (TCNT1_ICOccurred == 0 && !(TIFR1 & _BV(OCF1A)));  // wait for timeout or IC
   result = TCNT1_ICTimerValue;
   TCNT1_ICOccurred = 0;
   if (TIFR1 & _BV(OCF1A))    // if timeout return 0
      return 0;
   else if ( result == 0 )
      return 1;
   else
      return result;
}*/
uint16_t TCNT1_WaitForICOccurred( void )
{
   while (TCNT1_ICOccurred == 0);  // wait for IC
   TCNT1_ICOccurred = 0;
   return TCNT1_ICTimerValue;
}

/*!@brief Interrupt service routine for Timer/Counter1 Capture Event
 * 
 * Stores timer value at most recent event to TCNT1_ICTimerValue.  Toggles
 * edge source based on TCNT1_EdgeTypeToggle.  Sets TCNT1_ICOccurred to 1.
 */
ISR(TIMER1_CAPT_vect)
{
   // store timer value at input capture
   TCNT1_ICTimerValue = ICR1;

   // toggle edge select if flag was set in TCNT_InitIC()
   if ( TCNT1_EdgeTypeToggle ) {
      TCCR1B ^= (1<<ICES1);
   }

   // set status flag
   TCNT1_ICOccurred = 1;

   #ifdef TCNT1_TEST_IC
   // toggle PA0
   PORTA ^= (1<<PA0);
   #endif
}
