/*
   The Soma Project

   Copyright 2007 Rachel Pasetes, John Burr, Sean Eves Knudsen,
                  Thomas Gerdes, Andrew Parra, Erik Pasternak

  This file is part of The Soma Project.

  The Soma Project is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  The Soma Project is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with The Soma Project; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/

/*!@file usi.c
   @brief USI driver
   @ingroup tof
*/


#include <stdlib.h>
#include <avr/interrupt.h>
#include <assert.h>
#include "usi.h"
#include "../share/util.h"


/*  Speed configuration for master operation:
 *  Bits per second = CPUSPEED / PRESCALER / (COMPAREVALUE+1) / 2.
 *  Maximum = CPUSPEED / 64.
 */
#define TC0_PRESCALER_VALUE 8	//!< Must be 1, 8, 64, 256 or 1024.
#define TC0_COMPARE_VALUE   31	//!< Must be 0 to 255. Minimum 31 with prescaler CLK/1.


/*  Prescaler value converted to bit settings.
 */
#if TC0_PRESCALER_VALUE == 1
   #define TC0_PS_SETTING (1<<CS00)
#elif TC0_PRESCALER_VALUE == 8
   #define TC0_PS_SETTING (1<<CS01)
#elif TC0_PRESCALER_VALUE == 64
   #define TC0_PS_SETTING (1<<CS01)|(1<<CS00)
#elif TC0_PRESCALER_VALUE == 256
   #define TC0_PS_SETTING (1<<CS02)
#elif TC0_PRESCALER_VALUE == 1024
   #define TC0_PS_SETTING (1<<CS02)|(1<<CS00)
#else
   #error Invalid T/C0 prescaler setting.
#endif


/* USI port and pin definitions for ATtiny24/44/84.
 */
#define USI_OUT_REG	   PORTA //!< USI port output register.
#define USI_IN_REG	   PINA  //!< USI port input register.
#define USI_DIR_REG	   DDRA  //!< USI port direction register.
#define USI_CLOCK_PIN	PA4   //!< USI clock I/O pin.
#define USI_DATAIN_PIN	PA6   //!< USI data input pin.
#define USI_DATAOUT_PIN	PA5   //!< USI data output pin.

#define USI_SSDIR_REG   DDRA  //!< USI slave select direction register.
#define USI_SS_REG      PINA  //!< USI slave select register.
#define USI_SS_PIN      PA3   //!< USI slave select pin.

/* USI Buffer Defines */
#define USI_TX_BUFFER_SIZE 16  ///< Size of USI transmit buffer
#define USI_TX_BUFFER_MASK ( USI_TX_BUFFER_SIZE - 1 )
//!< Mask recieve buffer size for use in circular buffer
#if ( USART0_TX_BUFFER_SIZE & USART0_TX_BUFFER_MASK )
   #error TX buffer size is not a power of 2
#endif

/* Static Variables */
static uint8_t USI_TxBuf[USI_TX_BUFFER_SIZE];
static volatile uint8_t USI_TxBufCount;      ///< bytes waiting to tranfer, either in buffer or USIDR
static volatile uint8_t USI_TxHead;
static volatile uint8_t USI_TxTail;

static volatile uint8_t USI_TransmitComplete;
static volatile uint8_t USI_ReceivedData;
static volatile uint8_t USI_IsMaster;
static volatile uint8_t USI_IsActive;

//#define SLAVE_PA2_DEBUG ugh.

uint8_t USI_GetReceivedData( void )
{
   USI_TransmitComplete = 0;
   return USI_ReceivedData;
}

void USI_Transmit( uint8_t data )
{
   // enqueue byte if USI is not active or transmission in progress or buffer not empty
   if ( /*(!USI_IsActive) || */((USISR & 0x0F) != 0) || (USI_TxBufCount != 0)) {
      /* wait while buffer full */
      while ( USI_TxBufCount >= USI_TX_BUFFER_SIZE ); 
      USI_TxHead = ( USI_TxHead + 1 ) & USI_TX_BUFFER_MASK; 
      //USI_TxBuf[USI_TxHead] = (data>0x50)?(data|0x80):data;  // thom sucks.
      USI_TxBuf[USI_TxHead] = data;
      USI_TxBufCount++;
   } else {
      // not busy, nothing queued: ready data for next transmission
      //USI_TransmitComplete = 0;
      USIDR = data;
      USI_TxBufCount++;

      if ( USI_IsMaster ) {
         SET_BIT( TIFR0, OCF0A ); // clear compare match flag
         SET_BIT( TIMSK0, OCIE0A ); // enable compare match interrupt
      }
   }
}

void USI_ClearTransmitComplete( void )
{
   USI_TransmitComplete = 0;
}

void USI_WaitForTransmitComplete( void )
{
   while ( USI_TransmitComplete != 1 ) {};
   USI_TransmitComplete = 0;
}

void USI_WaitForBufferEmpty( void )
{
   while ( USI_TxBufCount != 0 );
}

/*!@brief ISR for Timer/Counter0 Compare Match: toggles USI clock output
   @note Only occurs on Master
 */
ISR(TIM0_COMPA_vect)
{
   SET_BIT( USICR, USITC); // toggle clock output pin
}

/*!@brief Interrupt service routine for USI Overflow (transmit complete).

   Stores new data in USIDR to USI_ReceivedData.  If there are 
   remaining bytes to send, a byte is dequeued for transmission.
*/
ISR(USI_OVF_vect)
{  
   if ( USI_IsMaster ) {
      CLR_BIT( TIMSK0, OCIE0A ); // disable compare match interrupt
   }

   USI_ReceivedData = USIDR;
   USI_TransmitComplete = 1;


   SET_BIT( USISR, USIOIF ); // clear overflow interrupt 

   // dequeue byte for next transmission if possible
   if ( (USI_IsActive==1) && ((USI_TxBufCount-1) > 0) ) {
      USI_TxTail = ( USI_TxTail + 1 ) & USI_TX_BUFFER_MASK;
      USIDR = USI_TxBuf[USI_TxTail];
      //USI_TransmitComplete = 0;
      if ( USI_IsMaster ) {
         SET_BIT( TIFR0, OCF0A ); // clear compare match flag
         SET_BIT( TIMSK0, OCIE0A ); // enable compare match interrupt
      }
   } else {
      // transmit NO_DATA if no meaningful data is ready
      USIDR = NO_DATA;
   }
   USI_TxBufCount--;
}


// should only be called if slave
ISR(PCINT0_vect)
{
   #ifdef SLAVE_PA2_DEBUG
   PORTA ^= (1<<PA2);
   #endif

   if ( /*!USI_IsActive &&*/ BIT_IS_CLR(USI_SS_REG,USI_SS_PIN) ) {
      // become active
      USI_IsActive = 1;
      //USI_TransmitComplete = 0;
      //USISR &= 0xF0; // set counter to zero
      SET_BIT( USISR, USIOIF );
      SET_BIT( USICR, USIWM0 );  // Three-wire mode
      
      SET_BIT( USI_DIR_REG, USI_DATAOUT_PIN );  // DO is driven as output

      if ( USI_TxBufCount == 0 ) {
         USIDR = NO_DATA;
      }
   } else if ( /*USI_IsActive &&*/ BIT_IS_SET(USI_SS_REG, USI_SS_PIN) ) {
      // stop transmitting at next transmission, 
      // even if data remains in buffer
      USI_IsActive = 0;
      CLR_BIT( USICR, USIWM0 );                 // Wire Mode unused (DO is GPIO)
      CLR_BIT( USI_DIR_REG, USI_DATAOUT_PIN );  // DO is Hi-Z 
      CLR_BIT( USI_OUT_REG, USI_DATAOUT_PIN );  // (input, not pulled up)
   } else {
      // ...shouldn't happen... it does, but everything seems to be working
      //assert( 0 );
   }
}

/*!@brief Initialize USI as SPI slave.
   
   Sets all necessary port directions and configures USI
   for SPI slave operation.

   @param spiMode Shift register clock source: 
                  0 for positive edge, 1 for negative edge.
   @note Interrupts must be enabled after this function is called.
*/
void USI_InitSPISlave( uint8_t spiMode )
{
   #ifdef SLAVE_PA2_DEBUG
   SET_BIT( DDRA, PA2 );
   SET_BIT( PORTA, PA2 );
   #endif

   /* set directions */
   CLR_BIT( USI_DIR_REG, USI_DATAOUT_PIN );  // DO is Hi-Z 
   CLR_BIT( USI_OUT_REG, USI_DATAOUT_PIN );  // (input, NOT pulled up)

   CLR_BIT( USI_DIR_REG, USI_DATAIN_PIN );   // DI is input
   SET_BIT( USI_OUT_REG, USI_DATAIN_PIN );   // DI is pulled high
   
   CLR_BIT( USI_DIR_REG, USI_CLOCK_PIN );    // USCK is input
   SET_BIT( USI_OUT_REG, USI_CLOCK_PIN );    // USCK is pulled high

   CLR_BIT( USI_SSDIR_REG, USI_SS_PIN );     // SS is input
   SET_BIT( USI_SS_REG, USI_SS_PIN );        // SS is pulled high 

   // SPI mode with counter overflow interrupt
   USICR = (1<<USIOIE) | /*(1<<USIWM0) | */ (1<<USICS1) | (spiMode<<USICS0);

   // Pin change interrupt 0 on PA3 on any edge
   SET_BIT( PCMSK0, PCINT3 );
   SET_BIT( GIMSK, PCIE0 );

   USISR &= 0xF0;
 
  // set initial status flags
   USI_IsMaster = 0;
   USI_IsActive = BIT_IS_CLR( USI_SS_REG, USI_SS_PIN );
   USICR |= (USI_IsActive<<USIWM0);
   USI_ReceivedData = 0;
   USI_TransmitComplete = 0;
   USI_TxHead = USI_TxTail = 0;
   USI_TxBufCount = 0;
}

/*!@brief Initialize USI as SPI master.
   
   Sets all necessary port directions and configures USI
   for SPI master operation.

   @param spiMode Shift register clock source:
                  0 for positive edge, 1 for negative edge.
   @note Interrupts must be enabled after this function is called.
*/
void USI_InitSPIMaster( uint8_t spiMode )
{
   /* set directions */
   SET_BIT( USI_DIR_REG, USI_DATAOUT_PIN );  // DO is output

   CLR_BIT( USI_DIR_REG, USI_DATAIN_PIN );   // DI is input
   SET_BIT( USI_OUT_REG, USI_DATAIN_PIN );   // DI is pulled high
   
   SET_BIT( USI_DIR_REG, USI_CLOCK_PIN );    // USCK is output  
   CLR_BIT( USI_OUT_REG, USI_CLOCK_PIN );    // USCK idles low

   // SPI mode with counter overflow interrupt
   USICR = (1<<USIOIE) | (1<<USIWM0) | 
           (1<<USICS1) | (spiMode<<USICS0) | (1<<USICLK);

   /* OC0A pin disconnected, must use compare match interrupt */
   CLR_BIT( TCCR0A, COM0A1 );
   CLR_BIT( TCCR0A, COM0A0 );
   
   /* Clear Timer on Compare Match mode */
   CLR_BIT( TCCR0B, WGM02 );
   SET_BIT( TCCR0A, WGM01 );
   CLR_BIT( TCCR0A, WGM00 );

   /* Set clock divide after prescaler based on #define */
   CLR_BIT( TCCR0B, CS02 );
   CLR_BIT( TCCR0B, CS01 );
   CLR_BIT( TCCR0B, CS00 );
   TCCR0B |= TC0_PS_SETTING;

   /* Init output compare register */
   OCR0A = TC0_COMPARE_VALUE;

   USISR &= 0xF0;

   // set initial status flags
   USI_IsMaster = 1;
   USI_IsActive = 1;
   USI_ReceivedData = 0;
   USI_TransmitComplete = 0;
   USI_TxHead = USI_TxTail = 0;
   USI_TxBufCount = 0;
}
